# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup
import logging
import requests

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import search_duckduckgo
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge
from komikku.webview import get_page_html

logger = logging.getLogger('komikku.servers.japscan')


class Japscan(Server):
    id = 'japscan'
    name = 'JapScan'
    lang = 'fr'
    long_strip_genres = ['Webtoon', ]
    status = 'disabled'

    has_cf = True

    base_url = 'https://www.japscan.lol'
    search_url = base_url + '/manga/'
    api_search_url = base_url + '/live-search/'
    manga_url = base_url + '/manga/{0}/'
    chapter_url = base_url + '/lecture-en-ligne/{0}/{1}/'
    page_url = '/lecture-en-ligne/{0}/{1}/{2}.html'
    cover_url = base_url + '{0}'
    bypass_cf_url = base_url + '/lecture-en-ligne/one-piece/1070/1.html'

    def __init__(self):
        self.session = None

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        return dict(slug=url.split('/')[-2])

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            chapters=[],
            server_id=self.id,
            synopsis=None,
        ))

        card_element = soup.find_all('div', class_='card')[0]

        # Main name: japscan handles several names for mangas (main + alternatives)
        # Name provided by search can be one of the alternatives
        # First word (Manga, Manhwa, ...) must be removed from name
        data['name'] = ' '.join(card_element.find('h1').text.strip().split()[1:])
        if data.get('cover') is None:
            data['cover'] = self.cover_url.format(card_element.find('img').get('src'))

        # Details
        if not card_element.find_all('div', class_='d-flex'):
            # mobile version
            elements = card_element.find_all('div', class_='row')[0].find_all('p')
        else:
            # desktop version
            elements = card_element.find_all('div', class_='d-flex')[0].find_all('p', class_='mb-2')

        for element in elements:
            label = element.span.text
            element.span.extract()
            value = element.text.strip()

            if label.startswith(('Auteur', 'Artiste')):
                for t in value.split(','):
                    t = t.strip()
                    if t not in data['authors']:
                        data['authors'].append(t)
            elif label.startswith('Genre'):
                data['genres'] = [genre.strip() for genre in value.split(',')]
            elif label.startswith('Statut'):
                # Possible values: ongoing, complete
                data['status'] = 'ongoing' if value == 'En Cours' else 'complete'

        # Synopsis
        synopsis_element = card_element.find('p', class_='list-group-item-primary')
        if synopsis_element:
            data['synopsis'] = synopsis_element.text.strip()

        # Chapters
        elements = soup.find('div', id='chapters_list').find_all('div', class_='chapters_list')
        for element in reversed(elements):
            if element.a.span:
                span = element.a.span.extract()
                # JapScan sometimes uploads some "spoiler preview" chapters, containing 2 or 3 untranslated pictures taken from a raw.
                # Sometimes they also upload full RAWs/US versions and replace them with a translation as soon as available.
                # Those have a span.badge "SPOILER", "RAW" or "VUS". We exclude these from the chapters list.
                if span.text.strip() in ('RAW', 'SPOILER', 'VUS', ):
                    continue

            slug = element.a.get('href').split('/')[3]

            data['chapters'].append(dict(
                slug=slug,
                title=element.a.text.strip(),
                date=convert_date_string(element.span.text.strip(), format='%d %b %Y'),
            ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url, decode=True):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages and scrambled are expected.
        """
        html = get_page_html(self.chapter_url.format(manga_slug, chapter_slug))
        soup = BeautifulSoup(html, 'lxml')

        if reader_element := soup.find(id='full-reader'):
            data = dict(
                pages=[],
            )

            img_elements = reader_element.find_all('img')
            if img_elements:
                # Full reader (several images)
                for img_element in img_elements:
                    data['pages'].append(dict(
                        url=None,
                        slug=None,
                        image=img_element.get('src'),
                    ))
            else:
                # Single reader (single image)
                for option_element in soup.find('select', id='pages').find_all('option'):
                    data['pages'].append(dict(
                        url=self.page_url.format(manga_slug, chapter_slug, int(option_element.get('value')) + 1),
                        slug=None,
                        image=None,
                    ))

            return data

        raise requests.exceptions.RequestException

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        if page['image']:
            # We already know the image URL
            url = page['image']
        elif page['url']:
            soup = BeautifulSoup(get_page_html(self.base_url + page['url']), 'lxml')
            if reader_element := soup.find(id='single-reader'):
                url = reader_element.img.get('src')
            else:
                return None

        r = self.session_get(
            url,
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=url.split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_latest_updates(self):
        """
        Returns recent manga
        """
        r = self.session_get(self.base_url, headers={'Referer': self.base_url})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('#chapters > div.tab-pane:not(:last-child) > .row > div > div.row > div > a:first-child'):
            results.append(dict(
                name=a_element.get('title').strip(),
                slug=a_element.get('href').split('/')[-2],
                cover=self.base_url + a_element.img.get('src') if a_element.img else None,
            ))

        return results

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns TOP manga
        """
        r = self.session_get(self.base_url, headers={'Referer': self.base_url})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for li_element in soup.find('div', id='top_mangas_all_time').find_all('li'):
            a_element = li_element.find_all('a')[0]
            results.append(dict(
                name=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-2],
            ))

        return results

    @CompleteChallenge()
    def search(self, term):
        r = self.session_post(self.api_search_url, data=dict(search=term), headers={
            'X-Requested-With': 'XMLHttpRequest',
            'Accept': '*/*',
            'Origin': self.base_url,
        })
        if r is None:
            return None

        if r.status_code == 200:
            try:
                data = r.json()

                results = []
                for item in data:
                    results.append(dict(
                        slug=item['url'].split('/')[-2],
                        name=item['name'],
                        cover=self.base_url + item['image'],
                    ))
            except Exception:
                pass  # noqa: TC202
            else:
                return results

        # Use DuckDuckGo Lite as fallback
        results = []
        for ddg_result in search_duckduckgo(self.search_url, term):
            # Remove first word in name (Manga, Manhua, Manhwa...)
            name = ' '.join(ddg_result['name'].split()[1:])
            # Keep only words before "|" character
            name = name.split('|')[0].strip()

            results.append(dict(
                name=name,
                slug=ddg_result['url'].split('/')[-2],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import os
import re
from urllib.parse import parse_qs
from urllib.parse import urlparse

import requests

from komikku.consts import USER_AGENT
from komikku.trackers import authenticated
from komikku.trackers import Tracker
from komikku.webview import get_tracker_access_token

# https://myanimelist.net/apiconfig/references/api/v2


class Myanimelist(Tracker):
    id = 'myanimelist'
    name = 'MyAnimeList'
    client_id = '4d0e78e295a090211517bded6ebdaa49'

    base_url = 'https://myanimelist.net'
    logo_url = base_url + '/images/favicon.ico'
    auth_url = base_url + '/v1/oauth2'
    authorize_url = auth_url + '/authorize?response_type=code&client_id={0}&state={1}&code_challenge={2}&code_challenge_method=plain'
    access_token_url = auth_url + '/token'

    api_url = 'https://api.myanimelist.net/v2'
    api_search_url = api_url + '/manga'
    api_manga_url = api_url + '/manga/{0}'
    api_manga_update_url = api_manga_url + '/my_list_status'
    api_user_mangalist_url = api_url + '/users/@me/mangalist'
    manga_url = base_url + '/manga/{0}'

    RELEASE_STATUSES = {
        'discontinued': 'Cancelled',
        'finished': 'Finished',
        'on_hiatus': 'Hiatus',
        'not_yet_published': 'Not Yet Released',
        'currently_publishing': 'Releasing',
    }

    STATUSES_MAPPING = {
        # tracker status => internal status
        'reading': 'reading',
        'completed': 'completed',
        'on_hold': 'on_hold',
        'dropped': 'dropped',
        'plan_to_read': 'plan_to_read',
        'rereading': 'rereading',  # Do not exists
    }

    USER_SCORE_FORMAT = {
        'min': 0,
        'max': 10,
        'step': 1,
        'raw_factor': 1,
    }

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_url(self, id):
        return self.manga_url.format(id)

    @authenticated
    def get_tracker_manga_data(self, id, access_token=None):
        r = self.session.get(
            self.api_manga_url.format(id),
            params={
                'fields': 'id,num_chapters,title',
            },
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        if r.status_code == 401:
            if self.refresh_access_token():
                return self.get_tracker_manga_data(id)

        elif r.status_code != 200:
            return None

        data = r.json()

        return {
            'id': data['id'],
            'name': data['title'],
            'chapters': data['num_chapters'],
            'score_format': None,
        }

    def get_user_score_format(self, format):
        return self.USER_SCORE_FORMAT

    @authenticated
    def get_user_manga_data(self, id, access_token=None):
        r = self.session_get(
            self.api_user_mangalist_url,
            params={
                'limit': 1000,
                'fields': 'id,my_list_status,num_chapters,title',
            },
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        if r.status_code == 401:
            if self.refresh_access_token():
                return self.get_user_manga_data(id)

        elif r.status_code != 200:
            return None

        data = None
        for item in r.json()['data']:
            if item['node']['id'] != id:
                continue

            if status := item['node']['my_list_status'].get('status'):
                status = self.STATUSES_MAPPING[status]
            else:
                status = 'reading'
            if status == 'reading' and item['node']['my_list_status']['is_rereading']:
                status = 'rereading'

            data = {
                'id': item['node']['id'],
                'name': item['node']['title'],
                'chapters': item['node']['num_chapters'],
                'chapters_progress': item['node']['my_list_status']['num_chapters_read'],
                'score': item['node']['my_list_status']['score'],
                'score_format': None,
                'status': status,
            }

        return data

    def refresh_access_token(self):
        r = self.session.post(
            self.access_token_url,
            auth=(self.client_id, ''),
            data={
                'grant_type': 'refresh_token',
                'refresh_token': self.data['refresh_token'],
            },
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
            }
        )

        if r.status_code != 200:
            return None

        data = r.json()

        self.data = {
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }

        return data['access_token']

    def request_access_token(self):
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode('utf-8')
        code_verifier = re.sub('[^a-zA-Z0-9]+', '', code_verifier)
        state = 'komikku'

        authorize_url = self.authorize_url.format(self.client_id, state, code_verifier)

        redirect_url, error = get_tracker_access_token(authorize_url, self.app_redirect_url)

        if error:
            return False, error

        qs = parse_qs(urlparse(redirect_url).query)
        if qs['state'][0] != state:
            return False, 'failed'

        code = qs['code'][0]

        r = self.session.post(
            self.access_token_url,
            data={
                'client_id': self.client_id,
                'grant_type': 'authorization_code',
                'code': code,
                'code_verifier': code_verifier,
            },
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
            }
        )

        if r.status_code != 200:
            return False, 'failed'

        data = r.json()
        self.data = {
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }

        return True, None

    @authenticated
    def search(self, term, access_token=None):
        r = self.session.get(
            self.api_search_url,
            params={
                'q': term,
                'limit': 100,
                'fields': 'id,title,main_picture,status,synopsis,genres,authors{first_name,last_name},mean,start_date',
            },
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        if r.status_code == 401:
            if self.refresh_access_token():
                return self.search(term)

        elif r.status_code != 200:
            return None

        results = []
        for item in r.json()['data']:
            authors = []
            if item['node'].get('authors'):
                for author in item['node']['authors']:
                    full = f'{author["node"]["first_name"]} {author["node"]["last_name"]}'.strip()
                    if author['role']:
                        full = f'{full} ({author["role"]})'
                    authors.append(full)

            results.append({
                'id': item['node']['id'],
                'authors': ', '.join(authors),
                'cover': item['node']['main_picture']['medium'] if item['node'].get('main_picture') else None,
                'name': item['node']['title'],
                'score': item['node']['mean'] if item['node'].get('mean') else None,
                'start_date': item['node']['start_date'][:4] if item['node'].get('start_date') else None,
                'status': self.RELEASE_STATUSES.get(item['node']['status']),
                'synopsis': item['node']['synopsis'],
            })

        return results

    @authenticated
    def update_user_manga_data(self, id, data, access_token=None):
        update_data = data.copy()

        # chapters_progress => num_chapters_read
        num_chapters_read = update_data.pop('chapters_progress', None)
        if num_chapters_read is not None:
            update_data['num_chapters_read'] = num_chapters_read

        if update_data.get('status') == 'rereading':
            update_data['status'] = 'reading'
            update_data['is_rereading'] = True

        update_data['status'] = self.convert_internal_status(update_data['status'])

        r = self.session_put(
            self.api_manga_update_url.format(id),
            data=update_data,
            headers={
                'Authorization': f'Bearer {access_token}',
            }
        )
        if r.status_code == 401:
            if self.refresh_access_token():
                return self.update_user_manga_data(id, data)

        elif r.status_code != 200:
            return False

        return True



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup
from datetime import datetime
import json
import re
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

#
# BEWARE: CatManga server is disabled
# Dead since 12/11/2021
#


class Catmanga(Server):
    id = 'catmanga'
    name = 'CatManga'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://catmanga.org'
    search_url = base_url + '/search'
    mangas_url = base_url
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/series/{0}/{1}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        json_data = json.loads(soup.find(
            'script',
            id=re.compile('__NEXT_DATA__'),
            type='application/json'
        ).string)

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        props = json_data['props']['pageProps']['series']
        data['name'] = props['title']
        data['cover'] = props['cover_art']['source']

        # Details
        data['authors'] = props['authors']
        data['genres'] = props['genres']
        if props['status'] == 'completed':
            data['status'] = 'complete'
        elif props['status'] == 'dropped':
            data['status'] = 'suspended'
        else:
            data['status'] = props['status']

        data['synopsis'] = props['description']

        # Chapters + scanlators
        for chapter in props['chapters']:
            if title := chapter.get('title'):
                title = f"{chapter['number']} - {title}"
            else:
                title = f"Chapter {chapter['number']}"

            if chapter.get('date'):
                date = convert_date_string(chapter.get('date'), format='%B %d, %Y')
            else:
                date = datetime.today().strftime('%Y-%m-%d')

            data['chapters'].append(dict(
                slug=str(chapter['number']),
                title=title,
                date=date,
                scanlators=chapter.get('groups'),
            ))
            data['scanlators'] = list(set(data['scanlators']) | set(chapter.get('groups')))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug)
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        json_data = json.loads(soup.find(
            'script',
            id=re.compile('__NEXT_DATA__'),
            type='application/json'
        ).string)
        pages = json_data['props']['pageProps']['pages']

        data = dict(
            pages=[],
        )
        for page in pages:
            data['pages'].append(dict(
                slug=None,
                image=page,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """ Returns chapter page scan (image) content """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """ Returns manga absolute URL """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """ Returns list of all mangas """
        r = self.session_get(self.mangas_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        json_data = json.loads(soup.find(
            'script',
            id=re.compile('__NEXT_DATA__'),
            type='application/json'
        ).string)

        series = json_data['props']['pageProps']['series']
        results = []
        for manga in series:
            results.append(dict(
                slug=manga['series_id'],
                name=manga['title'],
            ))
        return results

    def search(self, term):
        return list(filter(lambda x: term.lower() in x['name'].lower(), self.get_most_populars()))



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Webtoontr(Madara):
    id = 'webtoontr'
    name = 'Webtoon TR'
    lang = 'tr'
    status = 'disabled'  # Dead source
    is_nsfw = True

    date_format = '%d/%m/%Y'
    series_name = 'webtoon'

    base_url = 'https://webtoontr.net'
    logo_url = base_url + '/wp-content/uploads/2021/08/cropped-Icon2-32x32.png'
    chapter_url = base_url + '/' + series_name + '/{0}/{1}/'

    details_synopsis_selector = '.manga-excerpt'



# SPDX-FileCopyrightText: 2023-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# JManga [JA]
# MangaReader [EN/FR/JA/KO/ZH_HANS] (disabled)

from gettext import gettext as _
from io import BytesIO
import urllib.parse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.exceptions import ServerException
from komikku.servers.utils import unscramble_image_rc4
from komikku.utils import get_buffer_mime_type


class Mangareader(Server):
    base_url: str
    search_url: str
    list_url: str
    manga_url: str
    chapter_url: str
    api_chapter_images_url: str

    languages_codes: dict

    scrambled_images = False
    slug_position: int = -2

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(
            self.manga_url.format(initial_data['slug']),
            headers={
                'Referer': self.list_url,
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],  # not available
            'genres': [],
            'status': None,
            'cover': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
        })

        data['name'] = soup.select_one('.manga-name').text.strip()
        data['cover'] = soup.select_one('.manga-poster img').get('data-src')

        # Details
        for element in soup.select('.genres > a'):
            data['genres'].append(element.text.strip())

        for element in soup.select('.anisc-info .item'):
            label = element.span.text.strip()

            if label.startswith('タイプ'):
                # Type
                data['genres'].append(element.select_one('.name').text.strip())

            elif label.startswith('地位'):
                # Status
                value = element.select_one('.name').text.strip()
                if value == 'Publishing':
                    data['status'] = 'ongoing'
                elif value == 'Completed':
                    data['status'] = 'complete'
                elif value == 'Discontinued':
                    data['status'] = 'suspended'
                elif value == 'On Hiatus':
                    data['status'] = 'hiatus'

            elif label.startswith('著者'):
                # Authors
                for a_element in element.select('a'):
                    data['authors'].append(a_element.text.strip())

        if synopsis_element := soup.select_one('.description'):
            data['synopsis'] = synopsis_element.text.strip()

        # Chapters
        if ul_element := soup.select_one(f'#{self.languages_codes[self.lang]}-chaps'):
            for element in reversed(ul_element.select('li')):
                a_element = element.a
                sid = element.get('data-id')
                slug = a_element.get('href').split('/')[self.slug_position]
                data['chapters'].append({
                    'slug': f'{sid}:{slug}',
                    'title': a_element.get('title').strip(),
                    'num': element.get('data-number'),
                })
        else:
            # Manga exists but has no chapters in self.lang (not filtered in search)
            raise ServerException(_('Not available in {0} language').format(self.lang.upper()))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        chapter_id, chapter_slug = chapter_slug.split(':')

        # Get chapter images (ajax)
        r = self.session_get(
            self.api_chapter_images_url.format(chapter_id),
            headers={
                'Referer': urllib.parse.quote_plus(self.chapter_url.format(
                    manga_slug, self.languages_codes[self.lang], chapter_slug
                )),
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        json_data = r.json()
        if not json_data['status']:
            return None

        soup = BeautifulSoup(json_data['html'], 'lxml')

        data = {
            'pages': [],
        }
        for element in soup.select('.iv-card'):
            data['pages'].append({
                'slug': None,
                'scrambled': 'shuffled' in element.get('class'),
                'image': element.img.get('data-src'),
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.base_url + '/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        if self.scrambled_images and page['scrambled']:
            # js/read.min.js: key is 2nd argument of unShuffle function
            image = unscramble_image_rc4(r.content, 'stay', 200)
            with BytesIO() as io_buffer:
                image.save(io_buffer, 'png')
                buffer = io_buffer.getvalue()
        else:
            buffer = r.content

        return {
            'buffer': buffer,
            'mime_type': mime_type,
            'name': page['image'].split('?')[0].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, type=None, status=None, orderby='default'):
        params = {
            'type': type,
            'status': status,
            'language': self.languages_codes[self.lang],
            'sort': orderby,
        }

        r = self.session_get(
            self.list_url,
            params=params,
            headers={
                'Referer': self.list_url,
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for item in soup.select('.item-spc'):
            cover = None
            if cover_element := item.select_one('.manga-poster img'):
                cover = cover_element.get('data-src')
                if not cover:
                    cover = cover_element.get('src')

            results.append({
                'slug': item.select_one('.manga-name > a').get('href').split('/')[self.slug_position],
                'name': item.select_one('.manga-name').text.strip(),
                'cover': cover,
            })

        return results

    def get_latest_updates(self, type=None, status=None):
        return self.get_manga_list(type=type, status=status, orderby='latest-updated')

    def get_most_populars(self, type=None, status=None):
        return self.get_manga_list(type=type, status=status, orderby='most-viewed')

    def search(self, term, type=None, status=None):
        # Search does not take language into account
        r = self.session_get(
            self.search_url,
            params={
                'q': term,
            },
            headers={
                'Referer': self.base_url + '/',
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for item in soup.select('.item-spc'):
            langs = item.select_one('.tick-lang').text.strip().split('/')
            if self.languages_codes[self.lang].upper() not in langs:
                continue

            cover = None
            if cover_element := item.select_one('.manga-poster img'):
                cover = cover_element.get('data-src')
                if not cover:
                    cover = cover_element.get('src')

            results.append({
                'slug': item.select_one('.manga-name > a').get('href').split('/')[self.slug_position],
                'name': item.select_one('.manga-name').text.strip(),
                'cover': cover,
            })

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup

from komikku.servers.multi.manga_stream import MangaStream
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Carteldemanhwas(MangaStream):
    id = 'carteldemanhwas'
    name = 'Cartel De Manhwas'
    lang = 'es'
    is_nsfw = True

    chapters_order = 'desc'
    date_format: str = '%d %B, %Y'
    series_name = 'manga'

    base_url = 'https://carteldemanhwa.net'
    manga_list_url = base_url + '/cartel-de-manhwas/'

    authors_selector = '.infox .spe span:-soup-contains("Autor del Manhwa")'
    genres_selector = '[itemprop="genre"]'
    scanlators_selector = None
    status_selector = '.infox .spe span:-soup-contains("Estado")'
    synopsis_selector = '[itemprop="description"]'
    chapter_pages_selector = '.reader-area img'

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        # Name & cover
        data['name'] = soup.select_one(self.name_selector).text.strip()

        if element := soup.select_one(self.thumbnail_selector):
            data['cover'] = element.get('src')

        # Details
        for element in soup.select(self.authors_selector):
            element.b.extract()
            if author := element.text.strip():
                data['authors'].append(author)

        for element in soup.select(self.genres_selector):
            genre = element.text.strip()
            if genre and genre not in data['genres']:
                data['genres'].append(genre)

        if element := soup.select_one('.infox .spe span:-soup-contains("Tipo")'):
            element.b.extract()
            if type_ := element.text.strip():
                data['genres'].append(type_)

        if element := soup.select_one(self.status_selector):
            element.b.extract()
            if status := element.text.strip():
                if status == 'Activo':
                    data['status'] = 'ongoing'

        if element := soup.select_one(self.synopsis_selector):
            data['synopsis'] = element.text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(soup)

        return data

    def get_manga_chapters_data(self, soup):
        chapters = []

        li_elements = soup.select('#chapter_list ul li')
        if self.chapters_order == 'desc':
            li_elements = reversed(li_elements)

        for li_element in li_elements:
            a_element = li_element.select_one('.epsleft a')

            slug = a_element.get('href').split('/')[self.slug_position]
            num = slug.split('-')[-1]
            title = a_element.text.strip()
            if date_element := li_element.select_one('.epsleft .date'):
                date = convert_date_string(date_element.text.strip(), languages=[self.lang], format=self.date_format)
            else:
                date = None

            chapters.append(dict(
                slug=slug,
                title=title,
                num=num if is_number(num) else None,
                date=date,
            ))

        return chapters

    def get_manga_list(self, title=None, type=None, orderby=None):
        r = self.session_get(
            self.manga_list_url,
            params=dict(
                status='',
                type=type,
                order=orderby or '',
                title=title,
            )
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.relat article a'):
            name = a_element.get('title')
            cover_element = a_element.select_one('img')

            results.append(dict(
                slug=a_element.get('href').split('/')[self.slug_position],
                name=name,
                cover=cover_element.get('src') if cover_element else None,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Homepage: https://github.com/Suwayomi/Suwayomi-Server

import datetime
from gettext import gettext as _
import logging

from bs4 import BeautifulSoup
from requests.auth import HTTPBasicAuth

from komikku.servers import Server
from komikku.servers.utils import do_login
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

logger = logging.getLogger(__name__)

STATUSES = {
    'ONGOING': 'ongoing',
    'COMPLETED': 'complete',
    'PUBLISHING_FINISHED': 'complete',
    'CANCELLED': 'suspended',
    'ON_HIATUS': 'hiatus',
}


class Suwayomi(Server):
    id = 'suwayomi'
    name = 'Suwayomi'
    description = _('Manga reader server')
    lang = ''

    has_login = True
    sync = True

    base_url = None  # Customizable via the settings
    logo_url = 'https://raw.githubusercontent.com/Suwayomi/Suwayomi-Server/master/server/src/main/resources/icon/faviconlogo-128.png'

    headers = {
        'User-Agent': 'Komikku Suwayomi',
    }

    def __init__(self, username=None, password=None, address=None):
        if address:
            self.base_url = address

        if username and password:
            self.do_login(username, password)

    @property
    def api_base_url(self):
        return self.base_url + '/api/v1'

    @property
    def api_opds_base_url(self):
        return self.base_url + '/api/opds/v1.2'

    @property
    def api_graphql_url(self):
        return self.base_url + '/api/graphql'

    @property
    def api_chapters_url(self):
        return self.api_base_url + '/manga/{0}/chapters'

    @property
    def api_cover_url(self):
        return self.api_base_url + '/manga/{0}/thumbnail'

    @property
    def api_image_url(self):
        return self.api_base_url + '/manga/{0}/chapter/{1}/page/{2}'

    @property
    def api_manga_url(self):
        return self.api_base_url + '/manga/{0}'

    @property
    def api_search_url(self):
        return self.api_opds_base_url + '/mangas'

    @property
    def manga_url(self):
        return self.base_url + '/manga/{0}'

    @do_login
    def get_manga_data(self, initial_data):
        """
        Returns serie data using REST API

        Initial data should contain at least serie's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update({
            'name': resp_data['title'].strip(),
            'authors': [],
            'scanlators': [],  # not available
            'genres': resp_data.get('genre'),
            'status': STATUSES.get(resp_data['status']),
            'synopsis': resp_data['description'].strip() if resp_data.get('description') else None,
            'chapters': [],
            'server_id': self.id,
            'cover': self.base_url + resp_data.get('thumbnailUrl'),
        })

        for key in ('author', 'artist'):
            if authors := resp_data.get(key):
                for author in authors.split(','):
                    author = author.strip()
                    if author not in data['authors']:
                        data['authors'].append(author)

        if last_fetched_at := resp_data.get('lastFetchedAt'):
            data['last_update'] = datetime.datetime.fromtimestamp(last_fetched_at, tz=datetime.UTC)

        # Chapters
        r = self.session_get(self.api_chapters_url.format(data['slug']))
        if r.status_code != 200:
            return None

        for chapter in reversed(r.json()):
            # Store Id and index in slug
            # Id is used to retrieve chapter data and reading sync
            # Index is used to retrieve page images
            slug = f'{chapter["id"]}:{chapter["index"]}'  # noqa

            chapter_data = {
                'slug': slug,
                'title': chapter['name'],
                'num': chapter['chapterNumber'] if is_number(chapter['chapterNumber']) else None,
                'read': 1 if chapter['read'] else 0,
                'scanlators': [chapter['scanlator']] if chapter.get('scanlator') else None,
            }

            if upload_date := chapter.get('uploadDate'):
                chapter_data['date'] = datetime.datetime.fromtimestamp(int(upload_date / 1000)).date()

            if last_read_at := chapter.get('lastReadAt'):
                chapter_data['last_read'] = datetime.datetime.fromtimestamp(last_read_at, tz=datetime.UTC)

                if data.get('last_read') is None or chapter_data['last_read'] > data['last_read']:
                    data['last_read'] = chapter_data['last_read']

            if last_page_read := chapter.get('lastPageRead'):
                chapter_data['last_page_read_index'] = last_page_read

            data['chapters'].append(chapter_data)

        return data

    @do_login
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns serie chapter data using GraphQL API

        Currently, only pages are expected.
        """
        id, _index = chapter_slug.split(':')

        r = self.session_post(
            self.api_graphql_url,
            json={
                'operationName': 'GET_CHAPTER_PAGES_FETCH',
                'variables': {
                    'input': {
                        'chapterId': int(id),
                    },
                },
                'query': """mutation GET_CHAPTER_PAGES_FETCH($input: FetchChapterPagesInput!) {
                    fetchChapterPages(input: $input) {
                        chapter {
                            id
                            pageCount
                            __typename
                        }
                        pages
                        __typename
                    }
                }""",
            }
        )
        if r.status_code != 200:
            return None

        data = {
            'pages': [],
        }
        for page in r.json()['data']['fetchChapterPages']['pages']:
            data['pages'].append({
                'slug': page.split('/')[-1],
                'image': None,
            })

        return data

    @do_login
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        _id, index = chapter_slug.split(':')

        r = self.session_get(self.api_image_url.format(manga_slug, index, page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': '{0:04d}.{1}'.format(int(page['slug']), mime_type.split('/')[-1]),
        }

    @do_login
    def get_manga_url(self, slug, url):
        """
        Returns serie absolute URL
        """
        return self.manga_url.format(slug)

    @do_login
    def get_most_populars(self):
        return self.search()

    def login(self, username, password):
        # Check authentication
        try:
            r = self.session.get(
                self.base_url,
                auth=HTTPBasicAuth(username, password)
            )
            if r.status_code != 200:
                return False

        except Exception as error:
            logger.warning(error)
            return False

        # Add credentials in session headers
        self.session.headers['Authorization'] = r.request.headers['Authorization']

        return True

    @do_login
    def search(self, term=None):
        """
        Searches series using OPDS API
        """
        def get_page_entries(page):
            params = {
                'pageNumber': page,
            }
            if term:
                params['query'] = term

            r = self.session_get(
                self.api_search_url,
                params=params
            )
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'xml')

            return soup.select('entry'), soup.select_one('link[rel="next"]') is not None

        more = True
        page = 1
        results = []
        while more:
            entries, more = get_page_entries(page)
            if entries:
                for entry in entries:
                    slug = entry.id.text.strip().split(':')[-1]

                    results.append({
                        'name': entry.title.text.strip(),
                        'slug': slug,
                        'cover': self.api_cover_url.format(slug),
                    })
                page += 1

        return results

    @do_login
    def update_chapter_read_progress(self, data, manga_slug, manga_name, chapter_slug, chapter_url):
        id, _index = chapter_slug.split(':')

        r = self.session_post(
            self.api_graphql_url,
            json={
                'operationName': 'UPDATE_CHAPTERS',
                'variables': {
                    'input': {
                        'ids': [int(id)],
                        'patch': {
                            'lastPageRead': data['page'] - 1,
                        },
                    },
                    'getBookmarked': False,
                    'getRead': False,
                    'getLastPageRead': True,
                    'chapterIdsToDelete': [],
                    'deleteChapters': False,
                    'mangaId': -1,
                    'trackProgress': False,
                },
                'query': 'fragment TRACK_RECORD_BIND_FIELDS on TrackRecordType {\n  id\n  remoteId\n  trackerId\n  remoteUrl\n  title\n  status\n  lastChapterRead\n  totalChapters\n  score\n  displayScore\n  startDate\n  finishDate\n  private\n  __typename\n}\n\nmutation UPDATE_CHAPTERS($input: UpdateChaptersInput!, $getBookmarked: Boolean!, $getRead: Boolean!, $getLastPageRead: Boolean!, $chapterIdsToDelete: [Int!]!, $deleteChapters: Boolean!, $mangaId: Int!, $trackProgress: Boolean!) {\n  updateChapters(input: $input) {\n    chapters {\n      id\n      isBookmarked @include(if: $getBookmarked)\n      isRead @include(if: $getRead)\n      lastReadAt @include(if: $getRead)\n      lastPageRead @include(if: $getLastPageRead)\n      manga @include(if: $getRead) {\n        id\n        unreadCount\n        lastReadChapter {\n          id\n          __typename\n        }\n        latestReadChapter {\n          id\n          __typename\n        }\n        firstUnreadChapter {\n          id\n          __typename\n        }\n        __typename\n      }\n      manga @include(if: $getBookmarked) {\n        id\n        bookmarkCount\n        __typename\n      }\n      __typename\n    }\n    __typename\n  }\n  deleteDownloadedChapters(input: {ids: $chapterIdsToDelete}) @include(if: $deleteChapters) {\n    chapters {\n      id\n      isDownloaded\n      manga {\n        id\n        downloadCount\n        __typename\n      }\n      __typename\n    }\n    __typename\n  }\n  trackProgress(input: {mangaId: $mangaId}) @include(if: $trackProgress) {\n    trackRecords {\n      ...TRACK_RECORD_BIND_FIELDS\n      __typename\n    }\n    __typename\n  }\n}',
            }
        )

        return r.status_code == 200



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import textwrap
from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
from io import BytesIO
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number
from komikku.utils import ServerContent

# Conversion ISO_639-1 codes => server codes
LANGUAGES_CODES = {
    'ar': 'ar_JO',
    'de': 'de',
    'en': 'en',
    'es': 'es',
    'es_419': 'es_CO',
    'fr': 'fr',
    'it': 'it',
    'ja': 'jp',
    'ko': 'kr_KR',
    'nl': 'nl',
    'pl': 'pl',
    'pt': 'pt',
    'pt_BR': 'pt_BR',
    'ru': 'ru_RU',
    'tr': 'tr_TR',
    'zh_Hans': 'cn',
}


def get_comic_from_slug(slug):
    # A previous slug format included the language (for ex. dbm_en)
    comic = slug.split('_')[0]
    if comic == 'dbm':
        comic = 'page'

    return comic


class Dbmultiverse(Server):
    id = 'dbmultiverse'
    name = 'Dragon Ball Multiverse'
    lang = 'en'
    content = ServerContent(
        type=[_('Doujinshi'), _('Fan Webcomic')]
    )
    true_search = False

    base_url = 'https://www.dragonball-multiverse.com'
    logo_url = base_url + '/favicon.ico'
    most_populars_url = None
    manga_url = None
    chapter_url = None
    page_url = None
    cover_url = base_url + '/imgs/read/{0}.jpg'

    status_complete = 'finished'

    def __init__(self):
        self.most_populars_url = self.base_url + f'/{LANGUAGES_CODES[self.lang]}/read.html'
        self.manga_url = self.base_url + f'/{LANGUAGES_CODES[self.lang]}/chapters.html?comic={{0}}'
        self.chapter_url = self.base_url + f'/{LANGUAGES_CODES[self.lang]}/chapters.html?comic={{0}}&chapter={{1}}'
        self.page_url = self.base_url + f'/{LANGUAGES_CODES[self.lang]}/{{0}}-{{1}}.html'
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        comic = get_comic_from_slug(initial_data['slug'])

        r = self.session_get(self.manga_url.format(comic))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': ['Gogeta Jr', 'Asura', 'Salagir', 'Et al.'],
            'scanlators': [],
            'genres': ['Shōnen', 'Dōjinshi'],
            'status': 'ongoing',
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': self.cover_url.format(comic),
        })

        for element in soup.select('.dbm-tags span'):
            data['genres'].append(element.text.strip())

        if self.status_complete in soup.select_one('.cadrelect').text.strip():
            data['status'] = 'complete'

        data['synopsis'] = soup.select_one('.cadrelect > p').text.strip()

        # Chapters
        for element in soup.select('.chapter'):
            url = element.a.get('href')
            qs = parse_qs(urlparse(url).query)
            slug = qs['chapter'][0]

            chapter_data = {
                'slug': slug,
                'title': element.h4.text.strip(),
                'num': slug if is_number(slug) else None,
                'date': None,
                'pages': [],
            }

            data['chapters'].append(chapter_data)

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        comic = get_comic_from_slug(manga_slug)

        r = self.session_get(self.chapter_url.format(comic, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = {
            'pages': [],
        }
        for element in soup.select('.pageslist > a > img'):
            data['pages'].append({
                'slug': element.get('title'),
                'image': None,
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        comic = get_comic_from_slug(manga_slug)

        r = self.session_get(self.page_url.format(comic, page['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if element := soup.select_one('#balloonsimg'):
            if img_element := element.img:
                url = img_element.get('src')
                if not url:
                    url = img_element.get('style').split(';')[0].split(':')[1][4:-1]
            else:
                url = element.get('style').split('(')[1].split(')')[0]

            balloons = []
            for belement in element.select('.balloon'):
                balloon = {
                    'text': belement.text,
                }
                for style in belement.get('style').split(';'):
                    if not style:
                        continue
                    label, value = style.split(':')
                    if label in ('left', 'top', 'width'):
                        balloon[label] = float(value[:-2])
                    elif label == 'font-size':
                        balloon['font-ratio'] = float(value[:-1]) / 100

                balloons.append(balloon)

        elif celebrate_element := soup.select_one('.cadrelect'):
            # Special page to celebrate 1000/2000/... pages
            # return first contribution image
            url = celebrate_element.select_one('img').get('src')
            balloons = None
        else:
            return None

        r = self.session_get(self.base_url + url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        buffer = r.content
        format_ = mime_type.split('/')[-1]

        if balloons:
            with BytesIO(buffer) as io_buffer:
                image = Image.open(io_buffer).convert('RGB')
            draw = ImageDraw.Draw(image)
            font_base_size = 14
            font_ratio = 1
            font = None

            for balloon in balloons:
                if font is None or balloon.get('font-ratio', 1) != font_ratio:
                    font_ratio = balloon.get('font-ratio', 1)
                    font = ImageFont.truetype('DejaVuSansCondensed.ttf', size=font_base_size * font_ratio)

                max_width = 1
                language = self.lang.replace('_', '-')
                while max_width < len(balloon['text']):
                    if font.getlength(balloon['text'][:max_width + 1], language=language) > balloon['width']:
                        break
                    max_width += 1

                offset_y = balloon['top']
                for line in textwrap.wrap(balloon['text'], width=max_width):
                    offset_x = max(0, balloon['width'] - font.getlength(line)) / 2
                    draw.text(xy=(balloon['left'] + offset_x, offset_y), text=line, font=font, fill=(0, 0, 0))
                    offset_y += font_base_size * font_ratio

            with BytesIO() as io_buffer:
                image.save(io_buffer, format_)
                buffer = io_buffer.getvalue()

            image.close()

        return {
            'buffer': buffer,
            'mime_type': mime_type,
            'name': '{0}.{1}'.format(page['slug'], format_),
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(get_comic_from_slug(slug))

    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('#dbm-reads .dbm-read'):
            url = element.a.get('href')
            if param := parse_qs(urlparse(url).query).get('comic'):
                slug = param[0]
                if slug == 'page':
                    slug = 'dbm'

                results.append({
                    'slug': slug,
                    'name': element.h3.text.strip(),
                    'cover': self.cover_url.format(get_comic_from_slug(slug)),
                })

        return results

    def search(self, term=None):
        # This server does not have a true search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results


class Dbmultiverse_ar(Dbmultiverse):
    id = 'dbmultiverse_ar'
    lang = 'ar'

    status_complete = 'انتهت'


class Dbmultiverse_de(Dbmultiverse):
    id = 'dbmultiverse_de'
    lang = 'de'

    status_complete = 'beendet'


class Dbmultiverse_es(Dbmultiverse):
    id = 'dbmultiverse_es'
    lang = 'es'

    status_complete = 'terminado'


class Dbmultiverse_es_419(Dbmultiverse):
    id = 'dbmultiverse_es_419'
    lang = 'es_419'

    status_complete = 'terminado'


class Dbmultiverse_fr(Dbmultiverse):
    id = 'dbmultiverse_fr'
    lang = 'fr'

    status_complete = 'terminé'


class Dbmultiverse_it(Dbmultiverse):
    id = 'dbmultiverse_it'
    lang = 'it'

    status_complete = 'concluso'


class Dbmultiverse_ja(Dbmultiverse):
    id = 'dbmultiverse_ja'
    lang = 'ja'

    status_complete = 'finished'


class Dbmultiverse_ko(Dbmultiverse):
    id = 'dbmultiverse_ko'
    lang = 'ko'

    status_complete = 'finished'


class Dbmultiverse_nl(Dbmultiverse):
    id = 'dbmultiverse_nl'
    lang = 'nl'

    status_complete = 'finished'


class Dbmultiverse_pl(Dbmultiverse):
    id = 'dbmultiverse_pl'
    lang = 'pl'

    status_complete = 'finished'


class Dbmultiverse_pt(Dbmultiverse):
    id = 'dbmultiverse_pt'
    lang = 'pt'

    status_complete = 'finished'


class Dbmultiverse_pt_br(Dbmultiverse):
    id = 'dbmultiverse_pt_br'
    lang = 'pt_BR'

    status_complete = 'finalizada'


class Dbmultiverse_ru(Dbmultiverse):
    id = 'dbmultiverse_ru'
    lang = 'ru'

    status_complete = 'finished'


class Dbmultiverse_tr(Dbmultiverse):
    id = 'dbmultiverse_tr'
    lang = 'tr'

    status_complete = 'finished'


class Dbmultiverse_zh_hans(Dbmultiverse):
    id = 'dbmultiverse_zh_hans'
    lang = 'zh_Hans'

    status_complete = 'finished'



# SPDX-FileCopyrightText: 2025 Seth Falco
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Seth Falco <seth@falco.fun>

from gettext import gettext as _
import json

from bs4 import BeautifulSoup

from komikku.servers.multi.hiveworks import Hiveworks
from komikku.utils import ServerContent


class Nixofnothing(Hiveworks):
    id = 'nixofnothing'
    name = 'Nix of Nothing'
    lang = 'en'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')]
    )

    base_url = 'https://www.nixofnothing.com'
    logo_url = base_url + '/favicon.ico'
    donate_url = base_url
    cover_url = base_url + '/templates/nix2024/images/logo.png'

    def get_metadata(self, soup: BeautifulSoup):
        linked_data_str = soup.find('script', attrs={'type': 'application/ld+json'}).contents[0]
        linked_data = json.loads(linked_data_str)

        return {
            'authors': ['M. Lee Lunsford'],
            'synopsis': linked_data['about'],
        }



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import threading

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

import cairo

from komikku.reader.pager.image import KImage
from komikku.models import Settings
from komikku.utils import log_error_traceback


class Page(Gtk.Overlay):
    __gtype_name__ = 'Page'
    __gsignals__ = {
        'rendered': (GObject.SignalFlags.RUN_FIRST, None, (bool, bool, )),
    }

    def __init__(self, pager, chapter, index):
        super().__init__(hexpand=True, vexpand=True)

        self.pager = pager
        self.reader = pager.reader
        self.window = self.reader.window

        self.chapter = self.init_chapter = chapter
        self.data = None
        self.index = self.init_index = index
        self.path = None
        self.image = None
        self.retry_button = None

        self._status = None    # rendering, allocable, rendered, offlimit, disposed
        self.error = None      # connection error, server error, corrupt file error
        self.loadable = False  # loadable from disk or downloadable from server (chapter pages are known)
        self.obscured = True

        if self.reader.reading_mode != 'webtoon':
            self.zoomable = True
            self.scrolledwindow = Gtk.ScrolledWindow()
            self.scrolledwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            self.scrolledwindow.set_kinetic_scrolling(True)
            self.scrolledwindow.set_overlay_scrolling(True)

            self.set_child(self.scrolledwindow)
        else:
            self.zoomable = False

        # Activity indicator
        self.activity_indicator = Adw.Spinner(
            halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER, width_request=48, height_request=48, visible=False
        )
        self.add_overlay(self.activity_indicator)

        if self.reader.ocr_translator.enabled:
            # Creating a drawing area allowing to select an image portion containing text
            # in order to perform text recognition and translate it
            self.ocr_selection_da = Gtk.DrawingArea(hexpand=True, vexpand=True, can_target=False)
            self.ocr_selection_da.set_draw_func(self.draw_ocr_selection)
            self.reader.ocr_translator.bind_property(
                'active', self.ocr_selection_da, 'can-target',
                GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
            )
            self.add_overlay(self.ocr_selection_da)

            drag_gesture = Gtk.GestureDrag.new()
            drag_gesture.set_propagation_phase(Gtk.PropagationPhase.BUBBLE)
            drag_gesture.set_button(1)
            self.ocr_selection_da.add_controller(drag_gesture)
            drag_gesture.connect('drag-begin', self.on_ocr_selection_drag_begin)
            drag_gesture.connect('drag-update', self.on_ocr_selection_drag_update, False)
            drag_gesture.connect('drag-end', self.on_ocr_selection_drag_update, True)

            self.ocr_selection_surface = None
            self.ocr_selection_start_x = 0
            self.ocr_selection_start_y = 0

    @property
    def hscrollable(self):
        if self.zoomable:
            adj = self.scrolledwindow.get_hadjustment()
            return adj.props.upper > adj.props.page_size

        return False

    @property
    def scrollable(self):
        return self.hscrollable or self.vscrollable

    @GObject.Property(type=str)
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def vscrollable(self):
        if self.zoomable:
            adj = self.scrolledwindow.get_vadjustment()
            return adj.props.upper > adj.props.page_size

        return False

    def dispose(self):
        self.status = 'disposed'

        if self.image:
            self.image.dispose()
            self.image = None

        if self.get_parent().__class__.__name__ == 'KInfiniteCanvas':
            # Webtoon pager: unparent from KInfiniteCanvas
            self.unparent()
        else:
            # RTL/LTR/Vertical pager: remove from Adw.Carousel
            self.get_parent().remove(self)

    def draw_ocr_selection(self, da, ctx, width, height):
        if not self.ocr_selection_surface:
            return

        ctx.set_source_surface(self.ocr_selection_surface, 0, 0)
        ctx.paint()

    def on_button_retry_clicked(self, _button):
        self.chapter = self.init_chapter
        self.index = self.init_index

        self.retry_button.set_visible(False)

        self.render(retry=True)

    def on_clicked(self, _image, x, y):
        self.reader.pager.on_single_click(x, y)

    def on_ocr_selection_drag_begin(self, gesture, x, y):
        self.ocr_selection_start_x = x
        self.ocr_selection_start_y = y

    def on_ocr_selection_drag_update(self, gesture, offset_x, offset_y, end):
        self.ocr_selection_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.get_width(), self.get_height())

        ctx = cairo.Context(self.ocr_selection_surface)
        ctx.set_source_rgb(0.9647, 0.51, 0.462745)
        ctx.set_line_width(2)
        ctx.rectangle(self.ocr_selection_start_x, self.ocr_selection_start_y, offset_x, offset_y)
        ctx.stroke()

        self.ocr_selection_da.queue_draw()

        if end:
            # Clear context
            ctx.set_operator(cairo.OPERATOR_CLEAR)
            ctx.rectangle(0, 0, self.get_width(), self.get_height())
            ctx.paint()

            # Selection size
            w = offset_x
            h = offset_y
            if abs(w) < 1 or abs(h) < 1:
                return

            # Selection coordinates
            x = self.ocr_selection_start_x + min(0, offset_x)
            y = self.ocr_selection_start_y + min(0, offset_y)
            x = max(0, x - self.image.image_displayed_x)
            y = max(0, y - self.image.image_displayed_y)
            zoom = self.image.zoom

            # Text recognition
            self.reader.ocr_translator.recognize(
                self.image,
                x / zoom, y / zoom,
                (x + abs(w)) / zoom, (y + abs(h)) / zoom
            )

    def on_rendered(self, _image, update, retry):
        self.status = 'rendered'
        self.emit('rendered', update, retry)

    def on_zoom_begin(self, _image):
        self.reader.pager.interactive = False
        self.reader.toggle_controls(False)

    def on_zoom_end(self, _image):
        self.reader.pager.interactive = True

    def render(self, retry=False):
        def complete(error_code, error_message, stack_trace):
            if error_code in ('connection', 'server'):
                self.stop_activity_indicator()
                on_error(error_code, error_message, stack_trace)
                if retry:
                    return

            elif error_code == 'offlimit':
                self.stop_activity_indicator()
                self.status = 'offlimit'
                return

            if self.status == 'disposed':
                # Page has been removed from pager
                self.stop_activity_indicator()
                return

            self.set_image(retry)

        def load_chapter(prior_chapter=None):
            if self.chapter is None:
                return 'error', 'offlimit', None, None

            if self.chapter.pages and self.index >= 0 and self.index < len(self.chapter.pages):
                return 'success', None, None, None

            if self.index < 0:
                # Page belongs to another (previous) chapter
                self.chapter = self.reader.manga.get_next_chapter(self.chapter, -1)
                if self.chapter is None:
                    return 'error', 'offlimit', None, None

            if not self.chapter.pages:
                try:
                    if not self.chapter.update_full():
                        return 'error', 'server', None, None
                except Exception as e:
                    return 'error', 'connection', *log_error_traceback(e)

            if self.index > len(self.chapter.pages) - 1:
                # Page belongs to another (next) chapter
                prior_chapter = self.chapter
                self.chapter = self.reader.manga.get_next_chapter(self.chapter, 1)
                if self.chapter is None:
                    return 'error', 'offlimit', None, None

            if self.index < 0:
                self.index = len(self.chapter.pages) + self.index
            elif self.index > len(prior_chapter.pages if prior_chapter else self.chapter.pages) - 1:
                self.index = self.index - len(prior_chapter.pages)

            return load_chapter(prior_chapter)

        def on_error(kind, message=None, stack_trace=None):
            assert kind in ('connection', 'server', ), 'Invalid error kind'

            if message is not None:
                self.window.add_notification(message, timeout=2)

            elif stack_trace is not None and Settings.get_default().servers_bug_report:
                manga = self.reader.manga
                context = '\n'.join([
                    'Failed to fetch chapter data or page',
                    f'- Server ID: {manga.server.id}',
                    f'- Serie Name: {manga.name}',
                    f'- Serie URL: {manga.server.get_manga_url(manga.slug, manga.url)}',
                ])
                self.window.open_server_bug_report_dialog(manga.server, context, stack_trace)

            self.error = kind

            self.show_retry_button()

        def run():
            res, error_code, error_message, stack_trace = load_chapter()
            if res == 'error':
                GLib.idle_add(complete, error_code, error_message, stack_trace)
                return

            self.loadable = True

            if not self.reader.manga.is_local:
                page_path = self.chapter.get_page_path(self.index)
                if page_path is None:
                    try:
                        page_path, _rtime = self.chapter.get_page(self.index)
                        if page_path:
                            self.path = page_path
                        else:
                            error_code, error_message, stack_trace = 'server', None, None
                    except Exception as e:
                        error_code, error_message, stack_trace = 'connection', *log_error_traceback(e)
                else:
                    self.path = page_path
            else:
                try:
                    self.data = self.chapter.get_page_data(self.index)
                except Exception as e:
                    error_code, error_message, stack_trace = 'server', *log_error_traceback(e)

            GLib.idle_add(complete, error_code, error_message, stack_trace)

        if self.status is not None and self.error is None:
            return

        self.status = 'rendering'
        self.error = None

        self.start_activity_indicator()

        if not self.reader.manga.is_local:
            thread = threading.Thread(target=run)
            thread.daemon = True
            thread.start()
        else:
            run()

    def rescale(self):
        if self.image is None:
            return

        self.image.scaling = self.reader.scaling
        self.image.scaling_filter = self.reader.scaling_filter
        self.image.landscape_zoom = self.reader.landscape_zoom

    def set_allow_zooming(self, allow):
        if self.image is None:
            return

        if self.reader.reading_mode == 'webtoon':
            self.image.set_allow_zooming(False)
            return

        self.image.set_allow_zooming(allow)

    def set_image(self, retry):
        def on_loaded(image, error=None, mime_type=None):
            if error:
                self.error = error
                self.show_retry_button()

                if self.path:
                    GLib.unlink(self.path)

                message = [_('Failed to load image')]
                if self.error == 'corrupted-or-unsupported-format':
                    message.append(_('Corrupted file or unsupported image format ({0})').format(mime_type))
                self.window.add_notification('\n'.join(message), timeout=2)

            image.connect('clicked', self.on_clicked)
            image.connect('rendered', self.on_rendered, retry)
            image.connect('zoom-begin', self.on_zoom_begin)
            image.connect('zoom-end', self.on_zoom_end)

            if self.image:
                self.image.dispose()

            self.image = image
            self.status = 'allocable'
            if self.zoomable:
                self.scrolledwindow.set_child(self.image)
            else:
                self.set_child(self.image)

            self.stop_activity_indicator()

        if self.path is None and self.data is None:
            image = KImage()
            image.load_missing(callback=on_loaded)
        else:
            image = KImage(
                scaling=self.reader.scaling, scaling_filter=self.reader.scaling_filter,
                crop=self.reader.borders_crop, crop_threshold=self.reader.borders_crop_threshold,
                landscape_zoom=self.reader.landscape_zoom, zoomable=self.zoomable
            )
            image.load(path=self.path, data=self.data['buffer'] if self.data else None, callback=on_loaded)

    def show_retry_button(self):
        if self.retry_button is None:
            self.retry_button = Gtk.Button()
            self.retry_button.add_css_class('suggested-action')
            self.retry_button.set_valign(Gtk.Align.CENTER)
            self.retry_button.set_halign(Gtk.Align.CENTER)
            self.retry_button.connect('clicked', self.on_button_retry_clicked)

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
            icon = Gtk.Image.new_from_icon_name('view-refresh-symbolic')
            icon.set_icon_size(Gtk.IconSize.LARGE)
            vbox.append(icon)
            vbox.append(Gtk.Label(label=_('Retry')))
            self.retry_button.set_child(vbox)

            self.add_overlay(self.retry_button)

        self.retry_button.set_visible(True)

    def start_activity_indicator(self):
        if self.reader.reading_mode != 'webtoon':
            self.activity_indicator.set_visible(True)

    def stop_activity_indicator(self, *args):
        if self.reader.reading_mode != 'webtoon':
            self.activity_indicator.set_visible(False)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Mangaowl(Madara):
    id = 'mangaowl'
    name = 'Mangaowl'
    lang = 'en'
    is_nsfw = True

    series_name = 'read-1'  # This value changes regularly!

    base_url = 'https://mangaowl.io'
    logo_url = base_url + '/wp-content/uploads/2017/10/logo.png'
    chapters_url = base_url + '/read-1/{0}/ajax/chapters/'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent


class Monkeyuser(Server):
    id = 'monkeyuser'
    name = 'MonkeyUser'
    lang = 'en'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')]
    )
    true_search = False

    base_url = 'https://www.monkeyuser.com'
    donate_url = base_url
    logo_url = base_url + '/images/logo.png'
    chapter_url = base_url + '{0}'
    image_url = base_url + '{0}'
    cover_url = base_url + '/images/logo.png'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content
        """
        data = initial_data.copy()
        data.update(dict(
            authors=['Stefanache Cornel', 'Constantin Orasanu', 'Maria Sîrbu'],
            scanlators=[],
            genres=['Humor', 'Satire'],
            status='ongoing',
            synopsis='Software development satire in a web comic.',
            chapters=[],
            server_id=self.id,
            cover=self.cover_url,
        ))

        # Chapters
        for year in range(2016, datetime.date.today().year + 1):
            r = self.session_get(f'{self.base_url}/{year}/')
            if r.status_code != 200:
                continue

            mime_type = get_buffer_mime_type(r.content)
            if mime_type != 'text/html':
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            for element in reversed(soup.select('.comic')):
                a_element = element.select_one('a')
                title = a_element.text.split('|')[-1].strip()
                if 'animated' in title.lower():
                    continue

                data['chapters'].append(dict(
                    slug=a_element.get('href'),
                    date=convert_date_string(element.select_one('time').get('datetime')[:10], '%Y-%m-%d'),
                    title=title,
                ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if img_element := soup.select_one('.content img'):
            # Some chapters have no image but an embedded Youtube video
            return dict(
                pages=[
                    dict(
                        slug=None,
                        image=img_element.get('src'),
                    ),
                ]
            )

        return None

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.image_url.format(page['image']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.base_url

    def get_most_populars(self):
        return [dict(
            slug='',
            name='MonkeyUser',
            cover=self.cover_url,
        )]

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Hean CMS / Iken

# Supported servers:
# Aurora Scans [EN] (disabled)
# Hijala Scans [EN]
# Hive Toon [EN]
# Kenscans [EN]
# Mode Scanlator [pt_BR] (disabled)
# Reaper Scans [EN] (disabled)
# Reaper Scans [pt_BR] (disabled)
# Rezo Scans [EN] (disabled)


import logging
import time

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import parse_nextjs_hydration
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.webview import CompleteChallenge

logger = logging.getLogger('komikku.servers.multi.heancms')


class HeanCMS(Server):
    base_url: str
    api_url: str

    logo_url: str = None
    manga_url: str = None
    chapter_url: str = None
    api_manga_url: str = None
    api_chapters_url: str = None

    def __init__(self):
        if self.manga_url is None:
            self.manga_url = self.base_url + '/series/{0}'
        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/series/{0}/{1}'
        if self.api_manga_url is None:
            self.api_manga_url = self.api_url + '/post?postSlug={0}'
        if self.api_chapters_url is None:
            self.api_chapters_url = self.api_url + '/chapters?postId={0}&skip={1}&take={2}&order=desc&search=&userId=undefined'

        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data via API request

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()['post']

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['postTitle'],
            authors=[],
            scanlators=[],
            genres=[],
            status='ongoing',
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=resp_data['featuredImage'],
        ))

        for genre in resp_data['genres']:
            data['genres'].append(genre['name'])

        if resp_data['seriesStatus'] == 'COMPLETED':
            data['status'] = 'complete'
        elif resp_data['seriesStatus'] == 'ONGOING':
            data['status'] = 'ongoing'
        elif resp_data['seriesStatus'] == 'DROPPED':
            data['status'] = 'suspended'
        elif resp_data['seriesStatus'] == 'HIATUS':
            data['status'] = 'hiatus'

        if resp_data.get('author'):
            for author in resp_data['author'].split(','):
                data['authors'].append(author.strip())
        if resp_data.get('artist'):
            for artist in resp_data['artist'].split(','):
                data['authors'].append(artist.strip())

        if synopsis := resp_data['postContent']:
            data['synopsis'] = BeautifulSoup(synopsis, 'lxml').text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(resp_data['id'])

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Pages URLs are available in a <script> element
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if images := parse_nextjs_hydration(soup, 'images', key='images'):
            images = sorted(images, key=lambda m: m['order'])
            data = dict(
                pages=[],
            )
            for image in images:
                data['pages'].append(dict(
                    slug=None,
                    image=image['url'],
                ))

            return data

        return None

    def get_manga_chapters_data(self, serie_id):
        """
        Returns manga chapters list via API
        """
        chapters = []

        def get_page(serie_id, page):
            r = self.session_get(
                self.api_chapters_url.format(serie_id, (page - 1) * 50, page * 50),
                headers={
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return None, False, None

            data = r.json()
            if not data.get('post'):
                return None, False, None

            more = data['totalChapterCount'] > page * 50

            return data['post']['chapters'], more, get_response_elapsed(r)

        chapters = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            chapters_page, more, rtime = get_page(serie_id, page)
            if chapters_page:
                for chapter in chapters_page:
                    prefix = '🔒 ' if not chapter.get('isAccessible', True) else ''

                    chapters.append(dict(
                        slug=chapter['slug'],
                        title=f'{prefix}Chapter {chapter["number"]}',
                        num=chapter['number'],
                        date=convert_date_string(chapter['createdAt'].split('T')[0], '%Y-%m-%d') if 'createdAt' in chapter else None,
                    ))
                page += 1
                delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None

            elif chapters_page is None:
                # Failed to retrieve a chapters list page, abort
                break

        return list(reversed(chapters))

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    @CompleteChallenge()
    def get_latest_updates(self):
        """
        Returns latest updates
        """
        return self.get_manga_list(orderby='latest')

    def get_manga_list(self, term=None, orderby=None):
        params = dict(
            page=1,
            perPage=21,
        )
        if term:
            params['searchTerm'] = term
        else:
            if orderby == 'latest':
                params['orderBy'] = 'updatedAt'
            elif orderby == 'popular':
                params['orderBy'] = 'totalViews'

        r = self.session_get(
            self.api_url + '/query',
            params=params,
            headers={
                'Accept': 'application/json, text/plain, */*',
                'Origin': self.base_url,
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['posts']:
            cover = item['featuredImage']

            results.append(dict(
                slug=item['slug'],
                name=item['postTitle'],
                cover=cover,
                last_chapter=item['chapters'][0]['number'] if item.get('chapters') else None,
            ))

        return results

    @CompleteChallenge()
    def get_most_populars(self):
        """
        Returns most popular mangas
        """
        return self.get_manga_list(orderby='popular')

    @CompleteChallenge()
    def search(self, term):
        return self.get_manga_list(term=term)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from urllib.parse import urlsplit

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Scantrad France'

#
# BEWARE: Scantrad France server is disabled
# Dead since 03/2023
#


class Scantrad(Server):
    id = 'scantrad'
    name = SERVER_NAME
    lang = 'fr'
    status = 'disabled'

    base_url = 'https://scantrad.net'
    search_url = base_url
    most_populars_url = base_url + '/mangas'
    manga_url = base_url + '/{0}'
    chapter_url = base_url + '/mangas/{0}/{1}'
    image_url = 'https://scan-trad.fr/{0}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r is None:
            return None

        mime_type = get_buffer_mime_type(r.content)

        if r.status_code != 200 or mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[SERVER_NAME, ],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        container_element = soup.find('div', id='chap-top')
        title_element = container_element.find('div', class_='titre')
        data['name'] = get_soup_element_inner_text(title_element)
        data['cover'] = container_element.find('div', class_='ctt-img').img.get('src')

        # Details
        if authors_element := title_element.find('div', class_='titre-sub'):
            data['authors'] = authors_element.text.strip()[3:].split(', ')

        for element in container_element.find('div', class_='info').find_all('div', class_='sub-i'):
            label = get_soup_element_inner_text(element)

            if label.startswith('Genre'):
                data['genres'] = [span_element.text.strip() for span_element in element.find_all('span', class_='snm-button')]
            elif label.startswith('Status'):
                status = element.span.text.strip()
                if status == 'Arrêté':
                    data['status'] = 'suspended'
                elif status == 'En cours':
                    data['status'] = 'ongoing'
                elif status == 'Terminé':
                    data['status'] = 'complete'

        data['synopsis'] = container_element.find_all('div', class_='new-main')[0].p.text.strip()

        # Chapters
        if chapitres_container_element := soup.find('div', id='chapitres'):
            for element in reversed(chapitres_container_element.find_all('div', class_='chapitre')):
                data['chapters'].append(dict(
                    slug=element.a.get('href').split('/')[-1],
                    date=convert_date_string(element.find('div', class_='chl-date').text),
                    title='{0} {1}'.format(
                        element.find('span', class_='chl-num').text.strip(),
                        get_soup_element_inner_text(element.find('div', class_='chl-titre'))
                    ),
                ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        imgs_elements = soup.find('div', class_='main_img').find_all('img')

        data = dict(
            pages=[],
        )
        for img_element in imgs_elements:
            url = img_element.get('data-src')
            if not url or not url.startswith('lel'):
                continue

            data['pages'].append(dict(
                slug=None,
                image=url,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            self.image_url.format(page['image']),
            headers={
                'Referer': self.base_url + '/',
                'Host': urlsplit(self.image_url).netloc,
                'Accept': 'image/webp,*/*',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        r = self.session_get(self.most_populars_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.find_all('div', class_='manga'):
            a_element = element.find('div', class_='mr-info').a

            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=a_element.text.strip(),
            ))

        return results

    def search(self, term=None):
        r = self.session_post(self.search_url, data=dict(q=term))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.find_all('a'):
            name = a_element.find('div', class_='rgr-titre').text.strip()

            results.append(dict(
                slug=a_element.get('href').split('/')[-1],
                name=name,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Weebcentral(Server):
    id = 'weebcentral'
    name = 'Weeb Central'
    lang = 'en'

    base_url = 'https://weebcentral.com'
    logo_url = base_url + '/favicon.ico'
    search_url = base_url + '/search/data'
    manga_url = base_url + '/series/{0}'
    chapters_url = base_url + '/series/{0}/full-chapter-list'
    chapter_url = base_url + '/chapters/{0}'
    images_url = base_url + '/chapters/{0}/images?is_prev=False&reading_style=long_strip'
    cover_url = 'https://temp.compsci88.com/cover/normal/{0}.webp'

    filters = [
        {
            'key': 'adult',
            'type': 'select',
            'name': _('Adult Content'),
            'description': _('Filter by Adult Content'),
            'value_type': 'single',
            'default': 'Any',
            'options': [
                {'key': 'Any', 'name': _('Any')},
                {'key': 'True', 'name': _('Yes')},
                {'key': 'False', 'name': _('No')},
            ]
        },
        {
            'key': 'types',
            'type': 'select',
            'name': _('Types'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'Manga', 'name': _('Manga'), 'default': False},
                {'key': 'Manhwa', 'name': _('Manhwa'), 'default': False},
                {'key': 'Manhua', 'name': _('Manhua'), 'default': False},
                {'key': 'OEL', 'name': _('OEL'), 'default': False},
            ],
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Statuses'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'Ongoing', 'name': _('Ongoing'), 'default': False},
                {'key': 'Complete', 'name': _('Complete'), 'default': False},
                {'key': 'Hiatus', 'name': _('Hiatus'), 'default': False},
                {'key': 'Canceled', 'name': _('Canceled'), 'default': False},
            ]
        },
    ]
    long_strip_genres = ['Manhua', 'Manhwa']

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],  # not available
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
        })

        soup = BeautifulSoup(r.text, 'lxml')

        data['name'] = soup.select_one('h1').text.strip()
        data['cover'] = self.cover_url.format(data['slug'])

        # Details
        for a_element in soup.select('li:-soup-contains("Author") span a'):
            data['authors'].append(a_element.text.strip())

        for a_element in soup.select('li:-soup-contains("Tag") span a, li:-soup-contains("Type") a'):
            data['genres'].append(a_element.text.strip())

        if a_element := soup.select_one('li:-soup-contains("Status") a'):
            status = a_element.text.strip()

            if status == 'Complete':
                data['status'] = 'complete'
            elif status == 'Ongoing':
                data['status'] = 'ongoing'
            elif status == 'Hiatus':
                data['status'] = 'hiatus'
            elif status == 'Canceled':
                data['status'] = 'suspended'

        data['synopsis'] = soup.select_one('p.whitespace-pre-wrap.break-words').text.strip()

        # Chapters
        r = self.session_get(
            self.chapters_url.format(data['slug']),
            headers={
                'Hx-Current-Url': self.manga_url.format(data['slug']),
                'Hx-Request': 'true',
                'Hx-Target': 'chapter-list',
                'Referer': self.manga_url.format(data['slug']),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for a_element in reversed(soup.select('a')):
            title_element = a_element.select_one('span.flex > span')
            if not title_element:
                continue

            scanlator = 'Unknown'
            if img_element := soup.select_one('span:first-child > img'):
                if 'official' in img_element.get('src'):
                    scanlator = 'Offical'

            title = title_element.text.strip()
            num = title.split(' ')[-1]  # chapter number theoretically is at end of chapter title

            data['chapters'].append({
                'slug': a_element.get('href').split('/')[-1],
                'title': title,
                'scanlators': [scanlator],
                'num': num if is_number(num) else None,
                'date': convert_date_string(a_element.select_one('time').get('datetime').split('T')[0], '%Y-%m-%d'),
            })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(
            self.images_url.format(chapter_slug),
            headers={
                'Hx-Current-Url': self.chapter_url.format(chapter_slug),
                'Hx-Request': 'true',
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = {
            'pages': [],
        }
        for element in soup.select('img'):
            data['pages'].append({
                'slug': None,
                'image': element.get('src'),
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self, adult=None, types=None, statuses=None):
        return self.search(orderby='Latest Updates', adult=adult, statuses=statuses, types=types)

    def get_most_populars(self, adult=None, types=None, statuses=None):
        return self.search(orderby='Popularity', adult=adult, statuses=statuses, types=types)

    def search(self, term='', orderby='Best Match', adult=None, types=None, statuses=None):
        r = self.session_get(
            self.search_url,
            params={
                'author': '',
                'text': term,
                'sort': orderby,
                'order': 'Ascending',
                'official': 'Any',
                'adult': adult,
                'included_status': statuses,
                'included_type': types,
                'display_mode': 'Minimal Display',
            },
            headers={
                'Hx-Current-Url': f'{self.base_url}/search',
                'Hx-Request': 'true',
                'Hx-Target': 'search-results',
                'Hx-Trigger': 'advanced-search-form',
                'Referer': f'{self.base_url}/search',
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('article > a'):
            slug = a_element.get('href').split('/')[-2]

            results.append({
                'slug': slug,
                'name': a_element.text.strip(),
                'cover': self.cover_url.format(slug),
            })

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.genkan import Genkan
from komikku.servers.multi.genkan import GenkanInitial
from komikku.servers.multi.madara import Madara
from komikku.servers.multi.madara import Madara2


class Leviatanscans(Madara):
    id = 'leviatanscans'
    name = 'LSComic (LeviatanScans)'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://lscomic.com'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'

    details_status_selector = None
    details_synopsis_selector = '.manga-about'


class Leviatanscans_es(Madara2):
    id = 'leviatanscans_es'
    name = 'Escaneos de Leviatan'
    lang = 'es'
    status = 'disabled'

    base_url = 'https://es.leviatanscans.com'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'


class Leviatanscans__old(Genkan):
    id = 'leviatanscans__old'
    name = 'Leviatan Scans'
    lang = 'en'
    status = 'disabled'  # Switch to Madara (Wordpress)

    base_url = 'https://leviatanscans.com'
    search_url = base_url + '/comics?query={0}'
    most_populars_url = base_url + '/home'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/{1}'
    image_url = base_url + '{0}'


class Leviatanscans_es_old(GenkanInitial):
    id = 'leviatanscans_es_old'
    name = 'Leviatan Scans'
    lang = 'es'
    status = 'disabled'  # Switch to Madara (Wordpress)

    # Search is broken -> inherit from GenkanInitial instead of Genkan class

    base_url = 'https://es.leviatanscans.com'
    search_url = base_url + '/comics'
    most_populars_url = base_url + '/home'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/{1}'
    image_url = base_url + '{0}'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import base64
import datetime
from gettext import gettext as _
import time
from urllib.parse import urlparse

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number
from komikku.webview import CompleteChallenge

LANGUAGES_CODES = {
    'en': 'en',
    'es': 'es',
    'es_419': 'es-la',
    'fr': 'fr',
    'ja': 'ja',
    'pt': 'pt',
    'pt_BR': 'pt-br',
}
SEARCH_RESULTS_PAGES = 3
SEARCH_RESULTS_PAGE_LIMIT = 30
CHAPTERS_PAGE_LIMIT = 100

VRF_STAGES = [
    {
        'table': base64.b64decode('yINlmUNho8VYJT+ibTIP+9ESiULpVEtMOoD6U6lRE0R/xwXo/Xp9NrUgC4cw/Lmo33vUyjUE40kUoEWIr/fxfNNcq2s79ShQ5NhNrFnJ4hXPwOu/SuXzIbuTQKGFvfm08E9jvCfqAtoDqvQq3dVWPQFmJjgvkISBeXY3BgANR+yVnjGbcxZ47d6kLNfZPIayTq3/YGySb1KuVZodWp/WGNAO5pfMcpaK53Hhs0allBszaMaxuouOwdxbwgxIw6YunSsXjI05Yi0j9j4eHKfSXR8Ifo/Od+8iamRfCXTyvm7NGRGYdcQ0ywcK/u6RXhrbcCm4t2eCtrDgQVecJGkQ+A=='),
        'key': base64.b64decode('0Ec58JOY3uBzJK9m3zqIOpdlF7UFiax9DmA='),
        'iv': 0x5A,
    },
    {
        'table': base64.b64decode('IUFltCxD3Oc2cwCgkJffthaOg9cgPUb0LgW6H/VtfcF0kc5F25t+aWj6JH9VOhOaY0rAFdUxlDnl5BLNvwEJvQtP5qcw7vdb/K+chnbwnspSHT8mz5lqwz41TezG0hkO06FTjJZhsyNuFLDpD2ZZxQj/QIRcF90zpmQ7Byu483WsQqUE0C342HL+JXngRB6fRzxRyVTaKu83h7UYTJ0QMt6ixFh6S3F8gqkKwrGTL3jHNBsD45UnifK8+RGtishQV2K3rujLKEkiZxpr2dYcudFW4oFsDKhad3CLBvuyTqsCo4B7mL5IKQ1vXo/MOOvq1I1d8ar9X6Ttu5KF4fZgiA=='),
        'key': base64.b64decode('AAdjb1iPY8CiDmq9H34tKTBF8a3oDQ=='),
        'iv': 0x35,
    },
    {
        'table': base64.b64decode('NQHlu1/wVO5EmkwQymF810qqY2xG1k2obcas4Z9mCsPEIFl9pRIjFxbJ7ybMHbBckT5Ton85E0FOeHezbh/mjlEYpmpnlXOS8dgrqeq2KfxImTh1YK9y0PeMNhzA1OQzSY9brYOJq/l2QnE/hwOeZIhPixVSKIUlDb5vLcH6RWKxkIEMuP0bDwIqQ71AJJaEaMJL7A6YtyIwoRT+L5v4aZzodN/0+3nOGsfblFjgxSfPzVDjNFeNl5P26+kEC/8AHgdrpAbt3hHz3HrRN1Y6e+JHgF7ncFWnoF0y3THL1S71WgWGCa6KtSzTCCG58n68nTyj2T3Sshk7utqCtMi/ZQ=='),
        'key': base64.b64decode('DELOJgPsVaCcblDtTGMdHzM='),
        'iv': 0xBA,
    },
]


def convert_old_slug(slug):
    # Old format: slug.hid
    # New format: hid-slug
    if '.' in slug:
        slug, hid = slug.split('.')
        slug = f'{hid}:{slug}'

    return slug if ':' in slug else None


def compute_vrf(url, params):
    def encrypt_stage(data, table, key, iv):
        key_size = len(key)
        prev = iv

        out = b''
        for index, value in enumerate(data):
            prev = table[(value ^ key[index % key_size] ^ prev) & 0xFF] & 0xFF
            out += prev.to_bytes()

        return out

    new_params = []
    if params:
        params = dict(sorted(params.items()))
        for name in params.keys():
            if name.endswith('[]'):
                for index, value in enumerate(params[name]):
                    new_name = name.replace('[]', f'[{index}]')
                    new_params.append(f'{new_name}={value}')
            else:
                new_params.append(f'{name}={params[name]}')

    data = urlparse(url).path.replace('/api', '')
    if new_params:
        data += '?'
        data += '&'.join(new_params)

    data = data.encode()
    for stage in VRF_STAGES:
        data = encrypt_stage(data, stage['table'], stage['key'], stage['iv'])

    return base64.b64encode(data, altchars=b'-_').rstrip(b'=')


class Mangafire(Server):
    id = 'mangafire'
    name = 'MangaFire'
    lang = 'en'

    has_cf = True
    is_nsfw = True

    base_url = 'https://mangafire.to'
    logo_url = base_url + '/assets/mangafire/logo-sm.png'

    search_url = base_url + '/browse'
    manga_url = base_url + '/title/{0}'
    chapter_url = base_url + 'title/{0}/chapter/{1}'

    api_url = base_url + '/api'
    api_search_url = api_url + '/titles'
    api_manga_url = api_url + '/titles/{0}'  # hid
    api_chapters_url = api_url + '/titles/{0}/chapters'  # hid
    api_chapter_url = api_url + '/chapters/{0}'

    filters = [
        {
            'key': 'content_rating',
            'type': 'select',
            'name': _('Content Rating'),
            'description': _('Filter by Content Rating'),
            'value_type': 'multiple',
            'options': [
                {'key': 'safe', 'name': _('Safe'), 'default': True},
                {'key': 'suggestive', 'name': _('Suggestive'), 'default': True},
                {'key': 'erotica', 'name': _('Erotica'), 'default': False},
                {'key': 'pornographic', 'name': _('Pornographic'), 'default': False},
            ]
        },
        {
            'key': 'types',
            'type': 'select',
            'name': _('Types'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'manga', 'name': _('Manga'), 'default': False},
                {'key': 'manhwa', 'name': _('Manhwa'), 'default': False},
                {'key': 'manhua', 'name': _('Manhua'), 'default': False},
                {'key': 'other', 'name': _('Other'), 'default': False},
            ]
        },
        {
            'key': 'demographics',
            'type': 'select',
            'name': _('Demographics'),
            'description': _('Filter by Publication Demographics'),
            'value_type': 'multiple',
            'options': [
                {'key': 'josei', 'name': _('Josei'), 'default': False},
                {'key': 'seinen', 'name': _('Seinen'), 'default': False},
                {'key': 'shoujo', 'name': _('Shoujo'), 'default': False},
                {'key': 'shounen', 'name': _('Shounen'), 'default': False},
            ]
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Statuses'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'releasing', 'name': _('Ongoing'), 'default': False},
                {'key': 'finished', 'name': _('Completed'), 'default': False},
                {'key': 'on_hiatus', 'name': _('Hiatus'), 'default': False},
                {'key': 'discontinued', 'name': _('Canceled'), 'default': False},
            ]
        },
    ]

    long_strip_genres = [
        'manhwa',
    ]

    def __init__(self):
        self.session = None

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data using API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        if ':' not in initial_data['slug']:
            if slug := convert_old_slug(initial_data['slug']):
                initial_data['slug'] = slug
            else:
                # Invalid/obsolete slug
                return None

        hid, slug = initial_data['slug'].split(':')
        url = self.api_manga_url.format(hid)

        r = self.session_get(
            url,
            params={
                'vrf': compute_vrf(url, None),
            },
            headers={
                'Accept': 'application/json',
                'Referer': self.get_manga_url(initial_data['slug'], None),
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()['data']

        data = initial_data.copy()
        data.update({
            'name': resp_data['title'],
            'authors': [],
            'scanlators': [],  # not available?
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        if resp_data.get('poster') and resp_data['poster'].get('medium'):
            data['cover'] = resp_data['poster']['medium']

        for author in resp_data.get('authors', []):
            data['authors'].append(author['title'])
        for artist in resp_data.get('artists', []):
            artist = artist['title']
            if artist not in data['authors']:
                data['authors'].append(artist.strip())

        if resp_data['status'] == 'releasing':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'finished':
            data['status'] = 'complete'
        elif resp_data['status'] == 'on_hiatus':
            data['status'] = 'hiatus'
        elif resp_data['status'] == 'discontinued':
            data['status'] = 'suspended'

        for genre in resp_data.get('genres', []):
            data['genres'].append(genre['title'])
        for demographic in resp_data.get('demographics', []):
            data['genres'].append(demographic['title'])
        if type_ := resp_data.get('type'):
            data['genres'].append(type_)

        if synopsis := resp_data.get('synopsisHtml'):
            synopsis = synopsis.replace('<br>', '')
            synopsis = synopsis.replace('target="_blank"', '')
            synopsis = synopsis.replace('rel="noopener noreferrer"', '')
            data['synopsis'] = synopsis.strip()

        data['chapters'] = self.get_manga_chapters_data(data['slug'])

        return data

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns chapter data using API

        Currently, only pages are expected.
        """
        url = self.api_chapter_url.format(chapter_slug)

        r = self.session_get(
            url,
            params={
                'vrf': compute_vrf(url, None),
            },
            headers={
                'Accept': 'application/json',
                'Referer': self.chapter_url.format(manga_slug.replace(':', '-'), chapter_slug),
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        data = {
            'pages': [],
        }
        for index, page in enumerate(r.json()['data']['pages'], start=1):
            data['pages'].append({
                'slug': None,
                'image': page['url'],
                'index': index,
            })

        return data

    def get_manga_chapters_data(self, slug, page=1, chapters=None):
        if chapters is None:
            chapters = []

        hid, _slug = slug.split(':')
        url = self.api_chapters_url.format(hid)
        params = {
            'language': LANGUAGES_CODES[self.lang],
            'order': 'desc',
            'page': page,
            'limit': CHAPTERS_PAGE_LIMIT,
        }
        params['vrf'] = compute_vrf(url, params)

        r = self.session_get(
            url,
            params=params,
            headers={
                'Accept': 'application/json',
                'Referer': self.get_manga_url(slug, None),
                'X-Requested-With': 'XMLHttpRequest',
            }
        )
        if r.status_code != 200:
            return None

        rtime = get_response_elapsed(r)
        resp_data = r.json()

        for chapter in resp_data['items']:
            title = []
            if num := chapter.get('number'):
                title.append(f'Ch. {num}')
            if name := chapter.get('name'):
                title.append(name)
            type_ = chapter.get('type')

            chapters.append({
                'slug': chapter['id'],
                'title': ' '.join(title),
                'scanlators': [type_] if type_ else None,
                'num': num if is_number(num) else None,
                'date': datetime.datetime.fromtimestamp(chapter['createdAt']).date(),
            })

        if resp_data['meta']['hasNext']:
            if rtime:
                time.sleep(min(rtime * 4, DOWNLOAD_MAX_DELAY))

            self.get_manga_chapters_data(slug, page=page + 1, chapters=chapters)

        return list(reversed(chapters))

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': '{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        slug = slug.replace(':', '-')
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, content_rating=None, types=None, demographics=None, statuses=None, orderby=None):
        def get_page(page):
            params = {}
            if term:
                params['keyword'] = term

            if content_rating:
                params['content_rating[]'] = content_rating
            if types:
                params['types[]'] = types
            if demographics:
                params['demographics[]'] = demographics
            if statuses:
                params['statuses[]'] = statuses

            if orderby == 'popular':
                params['order[trending]'] = 'desc'
            elif orderby == 'latest':
                params['order[chapter_updated_at]'] = 'desc'
            else:
                params['order[relevance]'] = 'desc'

            params.update({
                'page': page,
                'limit': SEARCH_RESULTS_PAGE_LIMIT,
            })

            params['vrf'] = compute_vrf(self.api_search_url, params)

            r = self.session_get(
                self.api_search_url,
                params=params,
                headers={
                    'Accept': 'application/json',
                    'Referer': self.search_url,
                    'X-Requested-With': 'XMLHttpRequest',
                }
            )
            if r.status_code != 200:
                return [], False, None

            resp_data = r.json()

            more = resp_data['meta']['hasNext'] and page < SEARCH_RESULTS_PAGES

            return resp_data['items'], more, get_response_elapsed(r)

        results = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            for item in items:
                results.append({
                    'slug': f'{item["hid"]}:{item["slug"]}',  # noqa
                    'name': item['title'],
                    'cover': item['poster']['medium'],
                    'last_chapter': item.get('latestChapter'),
                })

            delay = min(rtime * 4, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    @CompleteChallenge()
    def get_latest_updates(self, content_rating=None, types=None, demographics=None, statuses=None):
        return self.get_manga_list(
            content_rating=content_rating, types=types, demographics=demographics, statuses=statuses, orderby='latest'
        )

    @CompleteChallenge()
    def get_most_populars(self, content_rating=None, types=None, demographics=None, statuses=None):
        return self.get_manga_list(
            content_rating=content_rating, types=types, demographics=demographics, statuses=statuses, orderby='popular'
        )

    @CompleteChallenge()
    def search(self, term, content_rating=None, types=None, demographics=None, statuses=None):
        return self.get_manga_list(
            term=term, content_rating=content_rating, types=types, demographics=demographics, statuses=statuses
        )


class Mangafire_es(Mangafire):
    id = 'mangafire_es'
    name = 'MangaFire'
    lang = 'es'


class Mangafire_es_419(Mangafire):
    id = 'mangafire_es_419'
    name = 'MangaFire'
    lang = 'es_419'


class Mangafire_fr(Mangafire):
    id = 'mangafire_fr'
    name = 'MangaFire'
    lang = 'fr'


class Mangafire_ja(Mangafire):
    id = 'mangafire_ja'
    name = 'MangaFire'
    lang = 'ja'


class Mangafire_pt(Mangafire):
    id = 'mangafire_pt'
    name = 'MangaFire'
    lang = 'pt'


class Mangafire_pt_br(Mangafire):
    id = 'mangafire_pt_br'
    name = 'MangaFire'
    lang = 'pt_BR'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Komga is a free and open source media server for your comics, mangas, BDs and magazines.
# Homepage: https://komga.org

from datetime import datetime
from gettext import gettext as _
import logging

from requests.auth import HTTPBasicAuth

from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import do_login
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

logger = logging.getLogger(__name__)


class Komga(Server):
    id = 'komga'
    name = 'Komga'
    description = _('Self-hosted digital library')
    lang = ''
    has_login = True
    sync = True

    base_url = None  # Customizable via the settings
    logo_url = 'https://komga.org/img/favicon.ico'

    headers = {
        'User-Agent': 'Komikku Komga',
        'Accept': 'application/json',
    }

    def __init__(self, username=None, password=None, address=None):
        if address:
            self.base_url = address

        if username and password:
            self.do_login(username, password)

    @property
    def api_base_url(self):
        return self.base_url + '/api/v1'

    @property
    def api_chapter_page_url(self):
        return self.api_base_url + '/books/{0}/pages/{1}'

    @property
    def api_chapter_pages_url(self):
        return self.api_base_url + '/books/{0}/pages'

    @property
    def api_chapter_read_progress(self):
        return self.api_base_url + '/books/{0}/read-progress'

    @property
    def api_chapters_url(self):
        return self.api_base_url + '/series/{0}/books?unpaged=true&media_status=READY'

    @property
    def api_cover_url(self):
        return self.api_base_url + '/series/{0}/thumbnail'

    @property
    def api_manga_url(self):
        return self.api_base_url + '/series/{0}'

    @property
    def api_search_url(self):
        return self.api_base_url + '/series'

    @property
    def manga_url(self):
        return self.base_url + '/series/{0}'

    @do_login
    def get_manga_data(self, initial_data):
        """
        Returns manga data using API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()
        metadata = resp_data['metadata']
        books_metadata = resp_data['booksMetadata']

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = metadata['title']
        data['cover'] = self.api_cover_url.format(data['slug'])

        # Details
        data['authors'] = [author['name'] for author in books_metadata['authors']]
        data['genres'] = [genre.capitalize() for genre in metadata['genres']]
        if metadata['status'] == 'ENDED':
            data['status'] = 'complete'
        elif metadata['status'] == 'ABANDONED':
            data['status'] = 'suspended'
        else:
            # Ongoing and hiatus
            data['status'] = metadata['status'].lower()
        data['reading_mode'] = metadata['readingDirection'].lower().replace('_', '-')

        data['synopsis'] = metadata['summary'] or books_metadata['summary']

        # Chapters
        r = self.session_get(self.api_chapters_url.format(data['slug']))
        if r.status_code != 200:
            return data

        items = r.json()['content']
        last_read = initial_data.get('last_read')  # only provided when updating
        for item in items:
            chapter_data = dict(
                slug=item['id'],
                title='#{0} {1}'.format(item['metadata']['number'], item['metadata']['title'].replace('_', ' ')),
                num=item['metadata']['number'] if is_number(item['metadata']['number']) else None,
                date=convert_date_string(item['metadata']['lastModified'].split('T')[0], format='%Y-%m-%d'),
            )

            if item.get('readProgress'):
                read_date = datetime.strptime(item['readProgress']['readDate'], '%Y-%m-%dT%H:%M:%SZ')
                if not last_read or read_date > last_read:
                    last_read = read_date
                    chapter_data.update(dict(
                        read=item['readProgress']['completed'],
                        last_read=read_date,
                        last_page_read_index=item['readProgress']['page'] - 1,
                    ))

            data['chapters'].append(chapter_data)

        data['last_read'] = last_read

        return data

    @do_login
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data

        Currently, only pages are expected.
        """
        r = self.session_get(self.api_chapter_pages_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        data = dict(
            pages=[],
        )
        for item in r.json():
            data['pages'].append(dict(
                slug=item['number'],
                image=item['fileName'].split('/')[-1],
            ))

        return data

    @do_login
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.api_chapter_page_url.format(chapter_slug, page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'],
        )

    @do_login
    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        return self.search('')

    def login(self, username, password):
        try:
            r = self.session.get(
                self.api_search_url,
                params={
                    'remember-me': True,
                },
                auth=HTTPBasicAuth(username, password)
            )
            if r.status_code != 200:
                return False
        except Exception as error:
            logger.warning(error)
            return False

        self.save_session()

        return True

    @do_login
    def search(self, term):
        r = self.session_get(self.api_search_url, params=dict(search=term))
        if r.status_code != 200:
            return None

        results = []
        for item in r.json()['content']:
            results.append(dict(
                name=item['name'],
                slug=item['id'],
                cover=self.api_cover_url.format(item['id']),
            ))

        return results

    @do_login
    def update_chapter_read_progress(self, data, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_patch(self.api_chapter_read_progress.format(chapter_slug), json=data)

        return r.status_code == 204



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import json
from urllib.parse import parse_qs
from urllib.parse import urlparse

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Rawkuma(Server):
    id = 'rawkuma'
    name = 'Rawkuma'
    lang = 'ja'
    is_nsfw = True

    base_url = 'https://rawkuma.net'
    logo_url = base_url + '/wp-content/uploads/2025/09/ラークマのサイトアイコンHEADER-96x96.png'
    manga_list_url = base_url + '/wp-admin/admin-ajax.php?action=advanced_search'
    manga_url = base_url + '/manga/{0}/'
    chapters_url = base_url + '/wp-admin/admin-ajax.php?action=chapter_list&manga_id={0}&page=1'
    chapter_url = base_url + '/manga/{0}/{1}/'
    nonce_url = base_url + '/wp-admin/admin-ajax.php?type=search_form&action=get_nonce'

    filters = [
        {
            'key': 'types',
            'type': 'select',
            'name': _('Types'),
            'description': _('Filter by Type'),
            'value_type': 'multiple',
            'options': [
                {'key': 'manga', 'name': _('Manga'), 'default': True},
                {'key': 'manhwa', 'name': _('Manhwa'), 'default': True},
                {'key': 'manhua', 'name': _('Manhua'), 'default': True},
                {'key': 'novel', 'name': _('Novel'), 'default': True},
            ],
        },
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],  # not available
            scanlators=[],  # not available
            genres=[],
            status=None,  # not available
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        # Name & cover
        data['name'] = soup.select_one('[itemprop="name"]').text.strip()
        data['cover'] = soup.select_one('[itemprop="image"] img').get('src')

        if elements := soup.select('[itemprop="genre"] span'):
            data['genres'] = [element.text.strip() for element in elements]

        if elements := soup.select('[itemprop="description"] p'):
            data['synopsis'] = [element.text.strip() for element in elements]
        if data['synopsis']:
            data['synopsis'] = '\n\n'.join(data['synopsis'])

        # Chapters
        if manga_id_element := soup.select_one('#chapter-list'):
            url = manga_id_element.get('hx-get').strip()
            manga_id = parse_qs(urlparse(url).query)['manga_id'][0]
        else:
            return None

        chapters = self.get_manga_chapters_data(manga_id)
        if chapters is not None:
            data['chapters'] = list(reversed(chapters))
        else:
            return None

        return data

    def get_manga_chapters_data(self, manga_id):
        r = self.session_get(self.chapters_url.format(manga_id))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        chapters = []
        for element in soup.select('[data-chapter-number]'):
            a_element = element.a

            num = element.get('data-chapter-number')

            title = a_element.select_one('div > div > div > span').text.strip()
            if date_element := element.select_one('time'):
                date = convert_date_string(date_element.get('datetime').split('T')[0], format='%Y-%m-%d')
            else:
                date = None

            chapters.append(dict(
                slug=a_element.get('href').split('/')[-2],
                title=title,
                num=num,
                date=date,
            ))

        return chapters

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            })
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        for element in soup.select('section > img'):
            data['pages'].append(dict(
                slug=None,
                image=element.get('src'),
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_list(self, title=None, types=None, orderby=None):
        r = self.session_post(
            self.manga_list_url,
            data={
                'nonce': self.get_nonce(),
                'inclusion': 'OR',
                'exclusion': 'OR',
                'page': 1,
                'genre': '[]',
                'genre_exclude': '[]',
                'author': '[]',
                'artist': '[]',
                'project': 0,
                'type': json.dumps(types) or '[]',
                'status': '[]',
                'order': 'desc',
                'orderby': orderby or 'title',
                'query': title or '',
            },
            headers={
                'Referer': f'{self.base_url}/library/',
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('div.rounded-lg'):
            a_element = element.a

            slug = a_element.get('href').split('/')[-2]
            name = a_element.img.get('alt')
            if span_element := element.select_one('div:-soup-contains("Chapter") > span'):
                last_chapter = span_element.text.strip()
            else:
                last_chapter = None

            results.append(dict(
                slug=slug,
                name=name,
                cover=a_element.img.get('src'),
                last_chapter=last_chapter,
            ))

        return results

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_nonce(self):
        r = self.session_get(self.nonce_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if input := soup.select_one('input'):
            return input.get('value')

        return None

    def get_latest_updates(self, types=None):
        return self.get_manga_list(types=types, orderby='updated')

    def get_most_populars(self, types=None):
        return self.get_manga_list(types=types, orderby='popular')

    def search(self, term, types=None):
        return self.get_manga_list(title=term, types=types)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.mangareader import Mangareader


class Mangareaderto(Mangareader):
    id = 'mangareaderto'
    name = 'MangaReader (to)'
    lang = 'en'
    is_nsfw = True
    status = 'disabled'

    languages_codes = dict(
        en='en',
        es='es',
        es_419='es-mx',
        fr='fr',
        ja='ja',
        ko='ko',
        pt='pt',
        pt_BR='pt-br',
        zh_Hans='zh',
    )

    base_url = 'https://mangareader.to'
    logo_url = base_url + '/favicon.ico?v=0.1'
    list_url = base_url + '/filter'
    search_url = base_url + '/search'
    manga_url = base_url + '/{0}'
    chapter_url = base_url + '/read/{0}/{1}/{2}'
    api_chapter_images_url = base_url + '/ajax/image/list/chap/{0}?mode=vertical&quality=medium&hozPageSize=1'


class Mangareaderto_es(Mangareaderto):
    id = 'mangareaderto_es'
    lang = 'es'


class Mangareaderto_es_419(Mangareaderto):
    id = 'mangareaderto_es_419'
    lang = 'es_419'


class Mangareaderto_fr(Mangareaderto):
    id = 'mangareaderto_fr'
    lang = 'fr'


class Mangareaderto_ja(Mangareaderto):
    id = 'mangareaderto_ja'
    lang = 'ja'


class Mangareaderto_ko(Mangareaderto):
    id = 'mangareaderto_ko'
    lang = 'ko'


class Mangareaderto_pt(Mangareaderto):
    id = 'mangareaderto_pt'
    lang = 'pt'


class Mangareaderto_pt_br(Mangareaderto):
    id = 'mangareaderto_pt_br'
    lang = 'pt_BR'


class Mangareaderto_zh_hans(Mangareaderto):
    id = 'mangareaderto_zh_hans'
    lang = 'zh_Hans'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import re

from komikku.servers.multi.madara import Madara2


class Manga3asq(Madara2):
    id = 'manga3asq'
    name = 'مانجا العاشق'
    lang = 'ar'

    date_format = '%Y \u060c%B %-d'

    base_url = 'https://3asq.online'
    logo_url = base_url + '/wp-content/uploads/2021/06/cropped-ICON-32x32.png'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'
    chapter_url = base_url + '/manga/{0}/{1}/'

    details_synopsis_selector = '.manga-excerpt'

    @staticmethod
    def extract_chapter_nums_from_slug(slug):
        re_nums = r'(\d+)[-_]?(\d+)?.*'

        if matches := re.search(re_nums, slug):
            if num := matches.group(1):
                num = f'{int(num)}'

                if num_dec := matches.group(2):
                    num = f'{num}.{int(num_dec)}'

                return num, None

        return None, None



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import json

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.multi.madara import Madara
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class Asurascans(Server):
    id = 'asurascans'
    name = 'Asura Scans'
    lang = 'en'

    base_url = 'https://asurascans.com'
    logo_url = base_url + '/images/logo.webp'
    search_url = base_url + '/browse'
    manga_url = base_url + '/comics/{0}'
    chapter_url = base_url + '/comics/{0}/chapter/{1}'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'manga', 'name': _('Manga')},
                {'key': 'manhwa', 'name': _('Manhwa')},
                {'key': 'manhua', 'name': _('Manhua')},
            ],
        },
    ]

    long_strip_genres = [
        'Manhwa',
        'Manhua',
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({
                'User-Agent': USER_AGENT,
            })

    def check_slug(self, initial_data):
        # A random number is always appended to slug and it changes regularly
        # Try to retrieve new slug
        res = self.search(initial_data['name'], '')
        if not res:
            return None

        for item in res:
            base_slug = '-'.join(initial_data['slug'].split('-')[:-1])
            current_base_slug = '-'.join(item['slug'].split('-')[:-1])
            if current_base_slug in (initial_data['slug'], base_slug) and initial_data['slug'] != item['slug']:
                return item['slug']

        return None

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        if new_slug := self.check_slug(initial_data):
            initial_data['slug'] = new_slug

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.content, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
        })

        # Name & cover
        data['name'] = soup.select_one('h1').text.strip()
        data['cover'] = soup.select_one('#mobile-cover-img').get('src')

        # Details
        for element in soup.select('a.inline-flex.text-xs.font-medium'):
            data['genres'].append(element.text.strip())

        if element := soup.select_one('div:-soup-contains("Type") ~ div span:last-child'):
            data['genres'].append(element.text.strip().capitalize())

        if element := soup.select_one('div:-soup-contains("Status") ~ div span:last-child'):
            status = element.text.strip().lower()
            if status in 'ongoing':
                data['status'] = 'ongoing'
            elif status == 'completed':
                data['status'] = 'complete'
            elif status in ('axed', 'dropped'):
                data['status'] = 'suspended'
            elif status == 'hiatus':
                data['status'] = 'hiatus'

        if author_element := soup.select_one('div:-soup-contains("Author") ~ a'):
            author = author_element.text.strip()
            if author and author != '_':
                data['authors'].append(author)
        if author_element := soup.select_one('div:-soup-contains("Artist") ~ a'):
            author = author_element.text.strip()
            if author and author != '_' and author not in data['authors']:
                data['authors'].append(author)

        if element := soup.select_one('#description-text'):
            data['synopsis'] = element.text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(soup)

        return data

    def get_manga_chapters_data(self, soup):
        chapters = []

        for a_element in reversed(soup.select('div.divide-y > a.group')):
            slug = a_element.get('href').split('/')[-1]
            if date_element := a_element.select_one('span.text-sm'):
                date = convert_date_string(date_element.text.strip())
            else:
                date = None

            chapters.append({
                'slug': slug,
                'title': a_element.select_one('span.font-medium').text.strip(),
                'num': slug,
                'date': date,
            })

        return chapters

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            })
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')
        info = json.loads(soup.select_one('astro-island').get('props'))

        data = {
            'pages': [],
        }
        for page in info['pages'][1]:
            data['pages'].append({
                'slug': None,
                'image': page[1]['url'][1],
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, type=None, orderby=None):
        params = {}
        if term:
            params['q'] = term
        if orderby:
            params['order'] = 'desc'
            if orderby == 'popular':
                params['sort'] = 'popular'
        if type:
            params['type'] = type

        r = self.session_get(
            self.search_url,
            params=params,
            headers={
                'Referer': self.base_url,
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.content, 'lxml')

        results = []
        for a_element in soup.select('.series-card[data-series-id] > a'):
            cover_element = a_element.select_one('img[loading="lazy"]')

            results.append({
                'slug': a_element.get('href').split('/')[-1],
                'name': cover_element.get('alt').strip(),
                'cover': cover_element.get('src') if cover_element else None,
            })

        return results

    def get_latest_updates(self, type):
        return self.get_manga_list(type=type, orderby='latest')

    def get_most_populars(self, type):
        return self.get_manga_list(type=type, orderby='popular')

    def search(self, term, type):
        return self.get_manga_list(term=term, type=type)


class Asurascans_tr(Madara):
    id = 'asurascans_tr'
    name = 'Armoni Scans (Asura Scans)'
    lang = 'tr'

    has_cf = True

    date_format = '%d %B %Y'

    base_url = 'https://asurascans.com.tr'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import html
import textwrap

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import TextImage
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent


class Fosscomics(Server):
    id = 'fosscomics'
    name = 'F/OSS Comics'
    lang = 'en'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')]
    )
    true_search = False

    base_url = 'https://fosscomics.com'
    donate_url = 'https://github.com/sponsors/joone'
    logo_url = base_url + '/images/favicon.png'
    manga_url = base_url + '/all_posts/'
    chapter_url = base_url + '/{0}'
    image_url = base_url + '/{0}/images/{1}'
    cover_url = base_url + '/8.%20The%20Origins%20of%20Unix%20and%20the%20C%20Language/images/feature.png'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping manga HTML page content
        """
        r = self.session_get(self.manga_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=['Joone Hur', ],
            scanlators=[],
            genres=[],
            status='ongoing',
            synopsis='Comics about Free and Open Source Software',
            chapters=[],
            server_id=self.id,
            cover=self.cover_url,
        ))

        # Chapters
        for element in reversed(soup.select('.posts .post')):
            slug = element.a.get('href').split('/')[-1]

            data['chapters'].append(dict(
                slug=slug,
                date=convert_date_string(element.span.text.strip(), '%a %b %d %Y'),
                title=element.a.text.strip(),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        i_index = 0  # images index
        p_index = 1  # paragraphs index
        for element in soup.select('section.body > p, section.body figure > img'):
            if element.name == 'p':
                # Text paragraph
                data['pages'].append(dict(
                    slug=None,
                    image=None,
                    text=html.unescape(element.text.strip()),
                    index=i_index,
                    subindex=p_index,
                ))
                p_index += 1
            else:
                # Image
                data['pages'].append(dict(
                    slug=None,
                    image=element.get('src').split('/')[-1],
                    index=i_index + 1,
                ))
                i_index += 1
                p_index = 1

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        chapter_num = int(chapter_slug.split('.')[0])

        if page.get('image'):
            r = self.session_get(self.image_url.format(chapter_slug, page['image']))
            if r.status_code != 200:
                return None

            mime_type = get_buffer_mime_type(r.content)
            if not mime_type.startswith('image'):
                return None

            name = f'{chapter_num:02d}_{page["index"]:02d}.{mime_type.split("/")[-1]}'  # noqa: E231
            content = r.content
        else:
            text = '\n'.join(textwrap.wrap(page['text'], 25))
            image = TextImage(text)

            mime_type = image.mime_type
            name = f'{chapter_num:02d}_{page["index"]:02d}_text_{page["subindex"]:02d}.{image.format}'  # noqa: E231
            content = image.content

        return dict(
            buffer=content,
            mime_type=mime_type,
            name=name,
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.base_url

    def get_most_populars(self):
        return [dict(
            slug='',
            name='F/OSS Comics',
            cover=self.cover_url,
        )]

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.foolslide import FoOlSlide


class Juinjutsuteam(FoOlSlide):
    id = 'juinjutsuteam'
    name = 'Juin Jutsu Team'
    lang = 'it'

    base_url = 'https://www.juinjutsureader.ovh'
    search_url = base_url + '/search/'
    latest_updates_url = base_url + '/latest/1/'
    manga_list_url = base_url + '/directory/'
    manga_url = base_url + '/series/{0}'
    chapter_url = base_url + '/read/{0}/it/{1}/page/1'

    name_selector = '#container_comic_content > .title_high'
    cover_selector = 'img.thumb'
    details_selector = '.info_comic b'
    synopsis_selector = '.trama'
    chapters_list_selector = '.list_chapter .element'
    chapters_list_link_selector = '.title_chapter > a'
    latest_updates_list_selector = '.list_elements > .group > .title_manga > a'
    search_list_selector = '.series_element'
    search_last_chapter_selector = '.cap a'
    search_list_cover_selector = 'img.thumb'



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from komikku.servers.multi.mangareader import Mangareader


class Jmanga(Mangareader):
    id = 'jmanga'
    name = 'JManga'
    lang = 'ja'
    is_nsfw = True

    base_url = 'https://jmanga.farm'
    list_url = base_url + '/filter/'
    search_url = base_url + '/'
    manga_url = base_url + '/read/{0}/'
    chapter_url = base_url + '/read/{0}/{1}/{2}/'
    api_chapter_images_url = base_url + '/json/chapter?mode=vertical&id={0}'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': 'all',
            'options': [
                {'key': 'all', 'name': '全て'},
                {'key': 'Raw Manga', 'name': 'Raw Manga'},
                {'key': 'BLコミック', 'name': 'BLコミック'},
                {'key': 'TLコミック', 'name': 'TLコミック'},
                {'key': 'オトナコミック', 'name': 'オトナコミック'},
                {'key': '女性マンガ', 'name': '女性マンガ'},
                {'key': '少女マンガ', 'name': '少女マンガ'},
                {'key': '少年マンガ', 'name': '少年マンガ'},
                {'key': '青年マンガ', 'name': '青年マンガ'},
            ],
        },
        {
            'key': 'status',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Status'),
            'value_type': 'single',
            'default': 'all',
            'options': [
                {'key': 'all', 'name': _('All')},
                {'key': 'Publishing', 'name': _('Ongoing')},
                {'key': 'Completed', 'name': _('Complete')},
            ],
        },
    ]

    languages_codes = {
        # 'en': 'en',
        'ja': 'ja',
    }



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.keyoapp import Keyoapp


class Kewnscans(Keyoapp):
    id = 'kewnscans'
    name = 'Kewn Scans'
    lang = 'en'

    base_url = 'https://kewnscans.org'
    logo_url = 'https://wsrv.nl/?url=cdn.meowing.org/uploads/d4846f15ded&w=20'

    genres_selector = 'div.justify-start > div.flex.gap-1 > a.flex'
